"""Test SectorModel and SectorModelBuilder
"""
from pytest import raises
from smif.parameters import ModelParameters
from smif.sector_model import SectorModel, SectorModelBuilder


class EmptySectorModel(SectorModel):
    def initialise(self, initial_conditions):
        pass

    def simulate(self, decisions, state, data):
        return state, {}

    def extract_obj(self, results):
        return 0


class TestSectorModelBuilder():

    def test_sector_model_builder(self, setup_project_folder):
        model_path = str(setup_project_folder.join('models', 'water_supply', '__init__.py'))
        builder = SectorModelBuilder('water_supply')
        builder.load_model(model_path, 'WaterSupplySectorModel')

        assets = [
            {
                'name': 'water_asset_a',
                'type': 'water_pump',
                'attributes': {
                    'capital_cost': 1000,
                    'economic_lifetime': 25,
                    'operational_lifetime': 25
                }
            }
        ]
        builder.add_interventions(assets)

        # builder.add_inputs(inputs)
        # builder.add_outputs(outputs)

        model = builder.finish()
        assert isinstance(model, SectorModel)

        assert model.name == 'water_supply'
        assert model.intervention_names == ['water_asset_a']
        assert model.interventions == assets

    def test_path_not_found(self):
        builder = SectorModelBuilder('water_supply')
        with raises(FileNotFoundError) as ex:
            builder.load_model('/fictional/path/to/model.py', 'WaterSupplySectorModel')
        msg = "Cannot find '/fictional/path/to/model.py' for the 'water_supply' model"
        assert msg in str(ex.value)

    def test_add_no_inputs(self, setup_project_folder):
        model_path = str(setup_project_folder.join('models', 'water_supply', '__init__.py'))
        builder = SectorModelBuilder('water_supply')
        builder.load_model(model_path, 'WaterSupplySectorModel')
        builder.add_inputs(None)
        assert isinstance(builder._sector_model.inputs, ModelParameters)
        assert len(builder._sector_model.inputs) == 0


class TestSectorModel(object):

    def test_interventions_names(self):
        assets = [
            {'name': 'water_asset_a'},
            {'name': 'water_asset_b'},
            {'name': 'water_asset_c'}
        ]
        model = EmptySectorModel()
        model.interventions = assets

        intervention_names = model.intervention_names

        assert len(intervention_names) == 3
        assert 'water_asset_a' in intervention_names
        assert 'water_asset_b' in intervention_names
        assert 'water_asset_c' in intervention_names

    def test_interventions(self):
        interventions = [
            {
                'name': 'water_asset_a',
                'capital_cost': 1000,
                'economic_lifetime': 25,
                'operational_lifetime': 25
            },
            {
                'name': 'water_asset_b',
                'capital_cost': 1500,
            },
            {
                'name': 'water_asset_c',
                'capital_cost': 3000,
            }
        ]
        model = EmptySectorModel()
        model.interventions = interventions
        actual = model.interventions

        assert actual == interventions

        assert sorted(model.intervention_names) == [
            'water_asset_a',
            'water_asset_b',
            'water_asset_c'
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Water Supply model
- WaterSupplySectorModel implements SectorModel
- wraps ExampleWaterSupplySimulationModel
- instantiate a model instance

"""

import logging
from smif import SpaceTimeValue, StateData
from smif.sector_model import SectorModel


class WaterSupplySectorModel(SectorModel):
    """Example of a class implementing the SectorModel interface,
    using one of the toy water models below to simulate the water supply
    system.
    """
    def initialise(self, initial_conditions):
        """Set up system here
        """
        pass

    def simulate(self, decisions, state, data):
        """

        Arguments
        =========
        decisions
            - asset build instructions, demand-side interventions to apply
        state
            - existing system/network (unless implementation means maintaining
              system in sector model)
            - system state, e.g. reservoir level at year start
        data
            - scenario data, e.g. expected level of rainfall
        """

        # unpack inputs
        reservoir_level = state[0].data['current_level']['value']
        raininess = sum([d.value for d in data['raininess']])

        # unpack assets
        number_of_treatment_plants = 2

        self.logger.debug(decisions)
        self.logger.debug(state)
        self.logger.debug(data)

        # simulate (wrapping toy model)
        instance = ExampleWaterSupplySimulationModel()
        instance.raininess = raininess
        instance.number_of_treatment_plants = number_of_treatment_plants
        instance.reservoir_level = reservoir_level

        water, cost = instance.run()
        results = {
            "water": [
                SpaceTimeValue('England', 1, water/3, "Ml"),
                SpaceTimeValue('Scotland', 1, water/3, "Ml"),
                SpaceTimeValue('Wales', 1, water/3, "Ml"),
            ],
            "cost": [
                SpaceTimeValue('England', 1, cost/3, "million £"),
                SpaceTimeValue('Scotland', 1, cost/3, "million £"),
                SpaceTimeValue('Wales', 1, cost/3, "million £"),
            ],
            "energy_demand": [
                SpaceTimeValue('England', 1, 3, "MWh"),
                SpaceTimeValue('Scotland', 1, 3, "MWh"),
                SpaceTimeValue('Wales', 1, 3, "MWh")
            ]
        }
        state = [
            StateData('Kielder Water', {'current_level': {'value': instance.reservoir_level}}),
        ]

        return state, results

    def extract_obj(self, results):
        return results['cost']

    def constraints(self, parameters):
        """

        Notes
        =====
        This constraint below expresses that water supply must be greater than
        or equal to 3.  ``x[0]`` is the decision variable for water treatment
        capacity, while the value ``parameters[0]`` in the min term is the
        value of the raininess parameter.
        """
        constraints = (
            {
                'type': 'ineq',
                'fun': lambda x: min(x[0], parameters[0]) - 3
            }
        )
        return constraints


class ExampleWaterSupplySimulationModel(object):
    """An example simulation model used for testing purposes

    Parameters
    ==========
    raininess : int
        The amount of rain produced in each simulation
    number_of_treatment_plants : int
        The amount of water is a function of the number of treatment plants and
        the amount of raininess
    """
    def __init__(self,
                 raininess=None,
                 number_of_treatment_plants=None,
                 reservoir_level=None):
        self.raininess = raininess
        self.number_of_treatment_plants = number_of_treatment_plants
        self.reservoir_level = reservoir_level

    def run(self):
        """Runs the water supply model

        Only 1 unit of water is produced per treatment plant,
        no matter how rainy.

        Each treatment plant costs 1.0 unit.
        """
        logger = logging.getLogger(__name__)

        logger.debug("There are %s plants", self.number_of_treatment_plants)
        logger.debug("It is %s rainy", self.raininess)

        logger.debug("Reservoir level was %s", self.reservoir_level)
        self.reservoir_level += self.raininess

        water = min(self.number_of_treatment_plants, self.reservoir_level)
        logger.debug("The system produces %s water", water)

        self.reservoir_level -= water
        logger.debug("Reservoir level now %s", self.reservoir_level)

        cost = 1.264 * self.number_of_treatment_plants
        logger.debug("The system costs £%s", cost)

        return water, cost


if __name__ == '__main__':
    """Run core model if this script is run from the command line
    """
    CORE_MODEL = ExampleWaterSupplySimulationModel(1, 1, 2)
    CORE_MODEL.run()



from pytest import fixture
from smif.decision import Planning


@fixture(scope='function')
def plan():
    planned_interventions = [
        {
            'name': 'small_pumping_station',
            'build_date': 2045,
            'location': 'oxford',
            'capacity': {
                'value': 500,
                'units': 'mcm/day'
            }
        },
        {
            'name': 'small_pumping_station',
            'build_date': 2035,
            'location': 'bicester',
            'capacity': {
                'value': 500,
                'units': 'mcm/day'
            }
        },
        {
            'name': 'large_pumping_station',
            'build_date': 2035,
            'location': 'abingdon',
            'capacity': {
                'value': 1500,
                'units': 'mcm/day'
            }
        }
    ]

    return Planning(planned_interventions)


class TestPlanning:

    def test_intervention_names(self, plan):
        expected = {'small_pumping_station', 'large_pumping_station'}
        actual = plan.names
        assert actual == expected

    def test_timeperiods(self, plan):
        expected = {2045, 2035}
        actual = plan.timeperiods
        assert actual == expected

    def test_empty(self):
        plan = Planning()
        assert plan.planned_interventions == []



# -*- coding: utf-8 -*-
"""A command line interface to the system of systems framework

This command line interface implements a number of methods.

- `setup` creates a new project folder structure in a location
- `run` performs a simulation of an individual sector model, or the whole
        system of systems model
- `validate` performs a validation check of the configuration file

Folder structure
----------------

When configuring a system-of-systems model for the CLI, the folder structure
below should be used.  In this example, there is one sector model, called
``water_supply``::

    /main_config.yaml
    /timesteps.yaml
    /water_supply.yaml
    /data/all/inputs.yaml
    /data/water_supply/
    /data/water_supply/inputs.yaml
    /data/water_supply/outputs.yaml
    /data/water_supply/assets/assets1.yaml
    /data/water_supply/planning/
    /data/water_supply/planning/pre-specified.yaml

The ``data`` folder contains one subfolder for each sector model.

The sector model implementations can be installed independently of the model
run configuration. The main_config.yaml file specifies which sector models
should run, while each set of sector model config

"""
from __future__ import print_function
import logging
import logging.config
import os
import re
import sys
import traceback
from argparse import ArgumentParser

import smif
from smif.sos_model import SosModelBuilder
from smif.data_layer.load import dump
from smif.data_layer.sos_model_config import SosModelReader
from smif.data_layer.sector_model_config import SectorModelReader
from smif.data_layer.validate import VALIDATION_ERRORS

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(name)-12s: %(levelname)-8s %(message)s'
        },
        'message': {
            'format': '\033[1;34m%(levelname)-8s\033[0m %(message)s'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.FileHandler',
            'level': 'DEBUG',
            'formatter': 'default',
            'filename': 'smif.log',
            'mode': 'a',
            'encoding': 'utf-8'
        },
        'stream': {
            'class': 'logging.StreamHandler',
            'formatter': 'message',
            'level': 'DEBUG'
        }
    },
    'root': {
        'handlers': ['file', 'stream'],
        'level': 'DEBUG'
    }
}

# Configure logging once, outside of any dependency on argparse
VERBOSITY = None
if '--verbose' in sys.argv:
    VERBOSITY = sys.argv.count('--verbose')
else:
    for arg in sys.argv:
        if re.match(r'\A-v+\Z', arg):
            VERBOSITY = len(arg) - 1
            break

if VERBOSITY is None:
    LOGGING_CONFIG['root']['level'] = logging.WARNING
elif VERBOSITY == 1:
    LOGGING_CONFIG['root']['level'] = logging.INFO
else:
    LOGGING_CONFIG['root']['level'] = logging.DEBUG

logging.config.dictConfig(LOGGING_CONFIG)
LOGGER = logging.getLogger(__name__)
LOGGER.debug('Debug logging enabled.')


def setup_project_folder(project_path):
    """Creates folder structure in the target directory

    Arguments
    =========
    project_path : str
        Absolute path to an empty folder

    """
    folder_list = ['data']
    for folder in folder_list:
        folder_path = os.path.join(project_path, folder)
        if os.path.exists(folder_path):
            msg = "{} already exists, skipping...".format(folder_path)
        else:
            msg = "Creating {} folder in {}".format(folder, project_path)
            os.mkdir(folder_path)
        LOGGER.info(msg)


def setup_configuration(args):
    """Sets up the configuration files into the defined project folder

    """
    project_path = os.path.abspath(args.path)
    msg = "Set up the project folders in {}?".format(project_path)
    response = confirm(msg,
                       response=False)
    if response:
        msg = "Setting up the project folders in {}".format(project_path)
        setup_project_folder(project_path)
    else:
        msg = "Setup cancelled."
    LOGGER.info(msg)


def run_model(args):
    """Runs the model specified in the args.model argument

    """
    model_config = validate_config(args)

    try:
        builder = SosModelBuilder()
        builder.construct(model_config)
        sos_model = builder.finish()
    except AssertionError as error:
        err_type, err_value, err_traceback = sys.exc_info()
        traceback.print_exception(err_type, err_value, err_traceback)
        err_msg = str(error)
        if len(err_msg) > 0:
            LOGGER.error("An AssertionError occurred (%s) see details above.", err_msg)
        else:
            LOGGER.error("An AssertionError occurred, see details above.")
        exit(-1)

    if args.model == 'all':
        LOGGER.info("Running the system of systems model")
        sos_model.run()
    else:
        LOGGER.info("Running the %s sector model", args.model)
        model_name = args.model
        sos_model.run_sector_model(model_name)

    output_file = args.output_file
    LOGGER.info("Writing results to %s", output_file)
    dump(sos_model.results, output_file)
    print("Model run complete")


def validate_config(args):
    """Validates the model configuration file against the schema

    Arguments
    =========
    args :
        Parser arguments

    """
    config_path = os.path.abspath(args.path)

    if not os.path.exists(config_path):
        LOGGER.error("The model configuration file '%s' was not found", config_path)
        exit(-1)
    else:
        try:
            # read system-of-systems config
            reader = SosModelReader(config_path)
            reader.load()

            model_config = reader.data
            config_basepath = os.path.dirname(config_path)

            # read sector model data+config
            model_config['sector_model_data'] = read_sector_model_data(
                config_basepath,
                model_config['sector_model_config'])
        except Exception as error:
            # should not raise error, so exit
            log_validation_errors()
            LOGGER.exception("Unexpected error validating config: %s", error)
            exit(-1)

        log_validation_errors()
        if len(VALIDATION_ERRORS) > 0:
            print("The model configuration was invalid")
            exit(-1)
        else:
            print("The model configuration was valid")

        return model_config


def log_validation_errors():
    """Log validation errors
    """
    for error in VALIDATION_ERRORS:
        LOGGER.error(error)


def path_to_abs(relative_root, path):
    """Return an absolute path, given a possibly-relative path
    and the relative root"""
    if os.path.isabs(path):
        return os.path.normpath(path)
    else:
        return os.path.normpath(os.path.join(relative_root, path))


def read_sector_model_data(config_basepath, config):
    """Read sector-specific data from the sector config folders
    """
    data = []

    for model_config in config:
        # read from dir relative to main model config file
        config_dir = path_to_abs(config_basepath, model_config['config_dir'])
        path = path_to_abs(config_basepath, model_config['path'])
        initial_conditions_paths = list(map(
            lambda path: path_to_abs(config_basepath, path),
            model_config['initial_conditions']
        ))
        interventions_paths = list(map(
            lambda path: path_to_abs(config_basepath, path),
            model_config['interventions']
        ))

        # read each sector model config+data
        reader = SectorModelReader({
            "model_name": model_config['name'],
            "model_path": path,
            "model_classname": model_config['classname'],
            "model_config_dir": config_dir,
            "initial_conditions": initial_conditions_paths,
            "interventions": interventions_paths
        })
        try:
            reader.load()
            data.append(reader.data)
        except FileNotFoundError as error:
            LOGGER.error("%s: %s", model_config['name'], error)
            raise ValueError("missing sector model configuration")

    return data


def parse_arguments():
    """Parse command line arguments

    Returns
    =======
    :class:`argparse.ArgumentParser`

    """
    parser = ArgumentParser(description='Command line tools for smif')
    parser.add_argument('-V', '--version',
                        action='version',
                        version="smif " + smif.__version__,
                        help='show the current version of smif')
    parser.add_argument('-v', '--verbose',
                        action='count',
                        help='show messages: -v to see messages reporting on progress, ' +
                        '-vv to see debug messages.')
    subparsers = parser.add_subparsers()

    # VALIDATE
    help_msg = 'Validate the model configuration file'
    parser_validate = subparsers.add_parser('validate',
                                            help=help_msg)
    parser_validate.set_defaults(func=validate_config)
    parser_validate.add_argument('path',
                                 help="Path to the main config file")

    # SETUP
    parser_setup = subparsers.add_parser('setup',
                                         help='Setup the project folder')
    parser_setup.set_defaults(func=setup_configuration)
    parser_setup.add_argument('path',
                              help="Path to the project folder")

    # RUN
    parser_run = subparsers.add_parser('run',
                                       help='Run a model')
    parser_run.add_argument('-m', '--model',
                            default='all',
                            help='The name of the model to run')
    parser_run.add_argument('-o', '--output-file',
                            default='results.yaml',
                            help='Output file')
    parser_run.set_defaults(func=run_model)
    parser_run.add_argument('path',
                            help="Path to the main config file")

    return parser


def confirm(prompt=None, response=False):
    """Prompts for a yes or no response from the user

    Arguments
    ---------
    prompt : str, default=None
    response : bool, default=False

    Returns
    -------
    bool
        True for yes and False for no.

    Notes
    -----

    `response` should be set to the default value assumed by the caller when
    user simply types ENTER.

    Examples
    --------

    >>> confirm(prompt='Create Directory?', response=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', response=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', response=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if response:
        prompt = '{} [{}]|{}: '.format(prompt, 'y', 'n')
    else:
        prompt = '{} [{}]|{}: '.format(prompt, 'n', 'y')

    while True:
        ans = input(prompt)
        if not ans:
            return response
        if ans not in ['y', 'Y', 'n', 'N']:
            print('please enter y or n.')
            continue
        if ans in ['y', 'Y']:
            return True
        if ans in ['n', 'N']:
            return False


def main(arguments=None):
    """Parse args and run
    """
    parser = parse_arguments()
    args = parser.parse_args(arguments)

    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""smif
"""
from __future__ import division, print_function, absolute_import

import pkg_resources

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


try:
    __version__ = pkg_resources.get_distribution(__name__).version
except:
    __version__ = 'unknown'


class SpaceTimeValue(object):
    """A piece of model input/output data

    Parameters
    ----------
    region: str
        A valid (unique) region name which is registered in the region register
    interval: str
        A valid (unique) interval name which is registered in the interval
        register
    value: float
        The value
    units: str
        The units associated with the `value`
    """
    def __init__(self, region, interval, value, units):
        self.region = region
        self.interval = interval
        self.value = value
        self.units = units

    def __repr__(self):
        return "SpaceTimeValue({}, {}, {}, {})".format(
            repr(self.region),
            repr(self.interval),
            repr(self.value),
            repr(self.units)
        )

    def __add__(self, other):
        if other.region != self.region:
            raise ValueError("Cannot add SpaceTimeValue of differing region")
        if other.interval != self.interval:
            raise ValueError("Cannot add SpaceTimeValue of differing interval")
        if other.units != self.units:
            raise ValueError("Cannot add SpaceTimeValue of differing units")

        return SpaceTimeValue(
            self.region,
            self.interval,
            self.value + other.value,
            self.units)

    def __eq__(self, other):
        return self.region == other.region and \
               self.interval == other.interval and \
               self.value == other.value and \
               self.units == other.units


class StateData(object):
    """A piece of state data

    Parameters
    ----------
    target
        The id or name of the object described by this state
    data
        The state attribute/data to apply - could typically be a dict of
        attributes
    """
    def __init__(self, target, data):
        self.target = target
        self.data = data

    def __repr__(self):
        return "StateData({}, {})".format(
            repr(self.target),
            repr(self.data)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for smif.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
from __future__ import absolute_import, division, print_function

import json
import logging

import yaml
from pytest import fixture

logging.basicConfig(filename='test_logs.log',
                    level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s: %(levelname)-8s %(message)s',
                    filemode='w')


@fixture(scope='function')
def one_dependency():
    """Returns a model input dictionary with a single (unlikely to be met)
    dependency
    """
    inputs = [
            {
                'name': 'macguffins produced',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

    return inputs


@fixture(scope='function')
def setup_folder_structure(tmpdir_factory):
    """

    Returns
    -------
    :class:`LocalPath`
        Path to the temporary folder
    """
    folder_list = ['config', 'data', 'models']
    test_folder = tmpdir_factory.mktemp("smif")

    for folder in folder_list:
        test_folder.mkdir(folder)

    return test_folder


@fixture(scope='function')
def setup_project_folder(setup_runpy_file,
                         setup_folder_structure,
                         setup_config_file,
                         setup_timesteps_file,
                         setup_water_inputs,
                         setup_water_outputs,
                         setup_time_intervals,
                         setup_regions,
                         setup_initial_conditions_file,
                         setup_pre_specified_planning,
                         setup_water_interventions_abc,
                         setup_interventions_file_one):
    """Sets up a temporary folder with the required project folder structure

        /config
        /config/model.yaml
        /config/timesteps.yaml
        /data
        /data/regions.geojson
        /data/intervals.yaml
        /data/water_supply/
        /data/water_supply/inputs.yaml
        /data/water_supply/outputs.yaml
        /data/water_supply/interventions/
        /data/water_supply/interventions/water_asset_abc.yaml
        /data/water_supply/interventions/assets_new.yaml
        /data/water_supply/pre-specified.yaml
        /models
        /models/water_supply/water_supply.py

    """
    base_folder = setup_folder_structure

    return base_folder


@fixture(scope='function')
def setup_project_missing_model_config(setup_runpy_file,
                                       setup_folder_structure,
                                       setup_config_file,
                                       setup_timesteps_file,
                                       setup_initial_conditions_file,
                                       setup_water_interventions_abc,
                                       setup_pre_specified_planning):
    """Sets up a temporary folder with the required project folder structure

        /config
        /config/model.yaml
        /config/timesteps.yaml
        /data
        /data/water_supply/
        /data/water_supply/pre-specified.yaml
        /data/water_supply/initial_conditions
        /data/water_supply/initial_conditions/assets_1.yaml
        /data/water_supply/interventions
        /data/water_supply/interventions/assets_1.yaml
        /models
        /models/water_supply/water_supply.py

        Deliberately missing:

        /data/water_supply/inputs.yaml
        /data/water_supply/outputs.yaml
        /data/intervals.yaml
        /data/water_supply/regions.geojson

    """
    base_folder = setup_folder_structure

    return base_folder


@fixture(scope='function')
def setup_project_empty_model_io(setup_runpy_file,
                                 setup_folder_structure,
                                 setup_config_file,
                                 setup_timesteps_file,
                                 setup_water_inputs,
                                 setup_water_outputs,
                                 setup_time_intervals,
                                 setup_regions,
                                 setup_initial_conditions_file,
                                 setup_pre_specified_planning,
                                 setup_water_interventions_abc,
                                 setup_interventions_file_one):
    """Sets up a temporary folder with the required project folder structure

        /config
        /config/model.yaml
        /config/timesteps.yaml
        /data
        /data/regions.geojson
        /data/intervals.yaml
        /data/water_supply/
        /data/water_supply/inputs.yaml
        /data/water_supply/outputs.yaml
        /data/water_supply/interventions/
        /data/water_supply/interventions/water_asset_abc.yaml
        /data/water_supply/interventions/assets_new.yaml
        /data/water_supply/pre-specified.yaml
        /models
        /models/water_supply/water_supply.py

    """
    base_folder = setup_folder_structure

    inputs_filename = base_folder.join('data', 'water_supply', 'inputs.yaml')
    inputs_filename.write('')

    outputs_filename = base_folder.join('data', 'water_supply', 'outputs.yaml')
    outputs_filename.write('')

    return base_folder


@fixture(scope='function')
def setup_initial_conditions_file(setup_folder_structure):
    """Assets are associated with sector models, not the integration config

    """
    base_folder = setup_folder_structure
    filename = base_folder.join('data',
                                'water_supply',
                                'initial_conditions',
                                'assets_1.yaml')
    assets_contents = [
        {
            'name': 'water_asset_a',
            'capacity': {
                'value': 5,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        },
        {
            'name': 'water_asset_b',
            'capacity': {
                'value': 15,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        },
        {
            'name': 'water_asset_c',
            'capacity': {
                'value': 25,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        }
    ]
    contents = yaml.dump(assets_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_initial_conditions_file_two(setup_folder_structure):
    """Assets are associated with sector models, not the integration config

    Defines a second assets file
    """
    base_folder = setup_folder_structure
    filename = base_folder.join('data',
                                'water_supply',
                                'initial_conditions',
                                'assets_2.yaml')
    assets_contents = [
        {
            'name': 'water_asset_d',
            'capacity': {
                'value': 15,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        }
    ]
    contents = yaml.dump(assets_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_interventions_file_one(setup_folder_structure):
    """Interventions are associated with sector models,
    not the integration config

    Defines an interventions file
    """
    base_folder = setup_folder_structure
    filename = base_folder.join('data',
                                'water_supply',
                                'interventions',
                                'assets_1.yaml')
    assets_contents = [
        {
            'name': 'water_asset_d',
            'capacity': {
                'value': 15,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford'
        }
    ]
    contents = yaml.dump(assets_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_interventions_file_two(setup_folder_structure):
    """Interventions are associated with sector models,
    not the integration config

    Defines an interventions file
    """
    base_folder = setup_folder_structure
    filename = base_folder.join('data',
                                'water_supply',
                                'interventions',
                                'assets_2.yaml')
    assets_contents = [
        {
            'name': 'water_asset_e',
            'capacity': {
                'value': 5,
                'units': 'Ml/day'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £"
            },
            'location': 'oxford'
        }
    ]
    contents = yaml.dump(assets_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_scenario_data(setup_folder_structure):
    file_contents = [
        {
            'value': 100,
            'units': 'people',
            'region': 'GB',
            'year': 2015
        },
        {
            'value': 150,
            'units': 'people',
            'region': 'GB',
            'year': 2016
        },
        {
            'value': 200,
            'units': 'people',
            'region': 'GB',
            'year': 2017
        }
    ]
    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('data', 'population.yaml')
    filepath.write(contents)

    # overwrite model.yaml
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/water_asset_abc.yaml'
                ]
            }
        ],
        'scenario_data': [
            {
                'parameter': 'population',
                'file': '../data/population.yaml',
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual'
            }
        ],
        'timesteps': 'timesteps.yaml',
        "region_sets": [{'name': 'national',
                         'file': 'regions.geojson'}],
        "interval_sets": [{'name': 'annual',
                           'file': 'intervals.yaml'}],
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': ['../data/water_supply/pre-specified.yaml']
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_no_planning(setup_folder_structure):
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/water_asset_abc.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {'use': False}
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_planning_missing(setup_folder_structure):
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/water_asset_abc.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': ['./does_not_exist.yaml']
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_planning_empty(setup_folder_structure):
    filepath = setup_folder_structure.join('data', 'water_supply', 'pre-specified.yaml')
    filepath.write('')


@fixture(scope='function')
def setup_abs_path_to_timesteps(setup_folder_structure):
    timesteps_abs_path = str(setup_folder_structure.join(
        'config',
        'timesteps.yaml'
    ))

    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/water_asset_abc.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': timesteps_abs_path,
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {'use': False}
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_config_file(setup_folder_structure):
    """Configuration file contains entries for sector models, timesteps and
    planning
    """
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/water_asset_abc.yaml'
                ]
            }
        ],
        'timesteps': 'timesteps.yaml',
        'scenario_data': [],
        "region_sets": [{'name': 'national',
                         'file': 'regions.geojson'}],
        "interval_sets": [{'name': 'annual',
                           'file': 'intervals.yaml'}],
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': ['../data/water_supply/pre-specified.yaml']
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_region_shapefile():
    'uk_nations_shp/regions.shp'


@fixture(scope='function')
def setup_pre_specified_planning_conflict(setup_folder_structure):
    """Sets up a configuration file for pre-specified planning

    """
    file_name = 'pre-specified_asset_d.yaml'
    file_contents = [
        {
            'name': 'water_asset_z',
            'description': 'Existing water treatment plants',
            'capacity': 6,
            'location': {'lat': 51.74556, 'lon': -1.240528},
            'build_date': 2015
        }
    ]
    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('data',
                                           'water_supply',
                                           file_name)
    filepath.write(contents)


@fixture(scope='function')
def setup_config_conflict_assets(setup_folder_structure,
                                 setup_pre_specified_planning_conflict):
    """Configuration file contains entries for sector models, timesteps and
    planning

    Contains conflicting assets in the pre-specified rules and the sector model
    """
    ps_name = 'pre-specified_asset_d.yaml'
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/assets1.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': [ps_name]
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_config_conflict_periods(setup_folder_structure,
                                  setup_timesteps_file_two):
    """Configuration file contains entries for sector models, timesteps and
    planning

    Contains conflicting assets in the pre-specified rules and the sector model
    """
    ps_name = 'pre-specified.yaml'
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/assets1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/assets1.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps2.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': [ps_name]
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_pre_specified_planning(setup_folder_structure):
    """Sets up a configuration file for pre-specified planning

    """
    file_name = 'pre-specified.yaml'
    file_contents = [
        {
            'name': 'water_asset_a',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_b',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_c',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        }
    ]
    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('data',
                                           'water_supply',
                                           file_name)
    filepath.write(contents, ensure=True)


@fixture(scope='function')
def setup_pre_specified_planning_two(setup_folder_structure):
    """Sets up a configuration file for pre-specified planning

    """
    file_name = 'pre-specified_alt.yaml'
    file_contents = [
        {
            'name': 'water_asset_a',
            'build_date': 2015,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_a',
            'build_date': 2020,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_a',
            'build_date': 2025,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        }
    ]
    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('data',
                                           'water_supply',
                                           file_name)
    filepath.write(contents)


@fixture(scope='function')
def setup_config_file_two(setup_folder_structure, setup_interventions_file_one):
    """Configuration file contains entries for sector models, timesteps and
    planning

    Contains two asset files
    """
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/initial_conditions/initial_2015_oxford.yaml'
                ],
                'interventions': [
                    '../data/water_supply/interventions/assets_1.yaml',
                    '../data/water_supply/interventions/assets_2.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': ['../data/water_supply/pre-specified.yaml']
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_config_file_timesteps_two(setup_folder_structure):
    """Configuration file contains entries for sector models, timesteps and
    planning
    """
    file_contents = {
        'sector_models': [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                'initial_conditions': [
                    '../data/water_supply/assets/assets_1.yaml'
                ],
                'interventions': [
                    '../data/water_supply/assets/assets_1.yaml'
                ]
            }
        ],
        'scenario_data': [],
        'timesteps': 'timesteps_2.yaml',
        'planning': {
            'rule_based': {'use': False},
            'optimisation': {'use': False},
            'pre_specified': {
                'use': True,
                'files': ['../data/water_supply/pre-specified.yaml']
            }
        }
    }

    contents = yaml.dump(file_contents)
    filepath = setup_folder_structure.join('config', 'model.yaml')
    filepath.write(contents)


@fixture(scope='function')
def setup_runpy_file(tmpdir, setup_folder_structure):
    """The python script should contain an instance of SectorModel which wraps
    the sector model and allows it to be run.
    """
    base_folder = setup_folder_structure
    # Write a file for the water_supply model
    filename = base_folder.join('models', 'water_supply', '__init__.py')
    contents = """
from smif.sector_model import SectorModel

class WaterSupplySectorModel(SectorModel):
    def initialise(self, initial_conditions):
        pass

    def simulate(self, decisions, state, data):
        pass

    def extract_obj(self, results):
        return 0

"""
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_water_inputs(setup_folder_structure):
    base_folder = setup_folder_structure
    filename = base_folder.join('data', 'water_supply', 'inputs.yaml')
    contents = [{'name': 'reservoir pumpiness',
                 'spatial_resolution': 'LSOA',
                 'temporal_resolution': 'annual'}]
    yaml_contents = yaml.dump(contents)
    filename.write(yaml_contents, ensure=True)


@fixture(scope='function')
def water_outputs_contents():
    contents = [
            {
                'name': 'storage_state',
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual'
            },
            {
                'name': 'storage_blobby',
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual'
            },
            {
                'name': 'total_water_demand',
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual'
            }
        ]
    return contents


@fixture(scope='function')
def setup_water_outputs(setup_folder_structure,
                        water_outputs_contents):
    base_folder = setup_folder_structure
    filename = base_folder.join('data', 'water_supply', 'outputs.yaml')
    contents = water_outputs_contents
    yaml_contents = yaml.dump(contents)
    filename.write(yaml_contents, ensure=True)
    return filename


@fixture(scope='function')
def setup_time_intervals(setup_folder_structure):
    base_folder = setup_folder_structure
    filename = base_folder.join('config', 'intervals.yaml')
    contents = [
        {
            "start": "P0Y",
            "end": "P1Y",
            "id": "whole_year"
        }
    ]
    yaml_contents = yaml.dump(contents)
    filename.write(yaml_contents, ensure=True)
    return filename


@fixture(scope='function')
def setup_region_data():
    data = {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:EPSG::27700"
            }
        },
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "oxford"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [448180, 209366],
                            [449500, 211092],
                            [450537, 211029],
                            [450873, 210673],
                            [451250, 210793],
                            [451642, 210023],
                            [453855, 208466],
                            [454585, 208468],
                            [456077, 207967],
                            [456146, 207738],
                            [456668, 207779],
                            [456708, 207444],
                            [456278, 207122],
                            [456149, 206615],
                            [455707, 206798],
                            [455749, 204521],
                            [456773, 204488],
                            [457014, 204184],
                            [456031, 203475],
                            [456444, 202854],
                            [456087, 202044],
                            [455369, 201799],
                            [454396, 202203],
                            [453843, 201634],
                            [452499, 203209],
                            [452052, 203566],
                            [451653, 203513],
                            [450645, 205137],
                            [449497, 205548],
                            [449051, 206042],
                            [448141, 208446],
                            [448180, 209366]
                        ]
                    ]
                }
            },
        ]
    }
    return data


@fixture(scope='function')
def setup_country_data():
    data = {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:EPSG::27700"
            }
        },
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "GB"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [0, 1],
                            [0, 1],
                            [2, 3]
                        ]
                    ]
                }
            },
            {
                "type": "Feature",
                "properties": {
                    "name": "NI"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [2, 3],
                            [2, 3],
                            [4, 5]
                        ]
                    ]
                }
            },
        ]
    }
    return data


@fixture(scope='function')
def setup_regions(setup_folder_structure, setup_region_data):
    base_folder = setup_folder_structure
    filename = base_folder.join('config', 'regions.geojson')
    data = setup_region_data
    filename.write(json.dumps(data), ensure=True)
    return filename


@fixture(scope='function')
def setup_results_file(setup_folder_structure):
    contents = \
        '''
Run Statistics
Software version: 111 System file version: 149
Saved on 29/11/2016 at 17:10

System file name:


System description:
   Run description:

------------------------------------------------------------
NLP solver: RELAX pure NLP
            Total number of NLP calls =     69241
            Total sim cpu (sec)    9.9360 Script cpu
------------------------------------------------------------
Mass balance report:
   Storage at start of replicate          =            216650
   + volume in arcs at start of replicate =                 0
   + total inflow                         =     6136254644472
   + total harvest node inflow            =                 0
   - total outflow at demand nodes        =         117263073
   - total flow into waste nodes          =     6136137397761
   - total reservoir evaporation          =                 0
   + backup reservoir evaporation         =                 0
   - total arc loss                       =                 0
 = Storage at end of replicate            =            200288
   + volume in arcs at end of replicate   =                 0
   + water balance error                  =                 0

TOTAL DEMAND                 =                         122427
  Node Demand node name                         Total shortfall
  -------------------------------------------------------------
    13 Test1                                               9080
    14 Test2                                               7832
    15 Test3                                               3876
------------------------------------------------------------
'''
    base_folder = setup_folder_structure
    # Write a results.txt file for the water_supply model
    filename = base_folder.join('models',
                                'water_supply',
                                'results.txt')
    filename.write(contents, ensure=True)
    return str(base_folder.join('models', 'water_supply'))


@fixture(scope='function')
def setup_timesteps_file(setup_folder_structure):
    base_folder = setup_folder_structure
    filename = base_folder.join('config', 'timesteps.yaml')
    timesteps_contents = [2010, 2011, 2012]
    contents = yaml.dump(timesteps_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_timesteps_file_two(setup_folder_structure):
    base_folder = setup_folder_structure
    filename = base_folder.join('config', 'timesteps_2.yaml')
    timesteps_contents = [2015, 2020, 2025]
    contents = yaml.dump(timesteps_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_timesteps_file_invalid(setup_folder_structure):
    base_folder = setup_folder_structure
    filename = base_folder.join('config', 'timesteps.yaml')
    timesteps_contents = "invalid timesteps file"
    contents = yaml.dump(timesteps_contents)
    filename.write(contents, ensure=True)


@fixture(scope='function')
def setup_water_interventions_abc(setup_folder_structure):
    data = [
        {
            "name": "water_asset_a",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_b",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1500
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_c",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 3000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        }

    ]
    content = yaml.dump(data)

    filename = setup_folder_structure.join('data',
                                           'water_supply',
                                           'interventions',
                                           'water_asset_abc.yaml')
    filename.write(content, ensure=True)

    return filename


@fixture(scope='function')
def setup_water_intervention_d(setup_folder_structure,
                               setup_config_file_two):

    content = """
- name: water_asset_d
  location: oxford
  operational_lifetime:
    units: years
    value: 50
  economic_lifetime:
    units: years
    value: 45
  capital_cost:
    units: "£"
    value: 3000
"""
    filename = setup_folder_structure.join('data',
                                           'water_supply',
                                           'interventions',
                                           'water_asset_d.yaml')
    filename.write(content, ensure=True)

    return filename


@fixture(scope='function')
def setup_minimal_water(setup_folder_structure,
                        setup_config_file):
    return str(setup_folder_structure)



# -*- coding: utf-8 -*-
"""Read and parse the config files for the system-of-systems model
"""

import logging
import os

import fiona

from .load import load
from .validate import (validate_scenario_data,
                       validate_sos_model_config,
                       validate_time_intervals,
                       validate_timesteps)


class SosModelReader(object):
    """Encapsulates the parsing of the system-of-systems configuration

    Parameters
    ----------
    config_file_path : str
        A path to the master config file

    """
    def __init__(self, config_file_path):
        self.logger = logging.getLogger(__name__)
        self.logger.info("Getting config file from %s", config_file_path)

        self._config_file_path = config_file_path
        self._config_file_dir = os.path.dirname(config_file_path)

        self._config = {}
        self.timesteps = []
        self.max_iterations = 0
        self.scenario_data = {}
        self.sector_model_data = []
        self.planning = []
        self.time_intervals = []
        self.regions = []

        self.names = []

        self.resolution_mapping = {
            'scenario': {},
            'input': {},
            'output': {}
        }

    def load(self):
        """Load and check all config
        """
        self._config = self.load_sos_config()
        self.timesteps = self.load_timesteps()
        self.max_iterations = self.load_max_iterations()
        self.time_intervals = self.load_time_intervals()
        self.regions = self.load_regions()
        self.scenario_data = self.load_scenario_data()
        self.sector_model_data = self.load_sector_model_data()
        self.planning = self.load_planning()

    @property
    def data(self):
        """Expose all model configuration data

        Returns
        -------
        dict
            Returns a dictionary with the following keys:

            timesteps
                the sequence of years
            max_iterations
                limit iterations for solving interdependencies
            sector_model_config: list
                The list of sector model configuration data
            scenario_data: dict
                A dictionary of scenario data, with the parameter name
                as the key and the data as the value
            planning: list
                A list of dicts of planning instructions
            region_sets: dict
                A dictionary of region set data, with the name as the key
                and the data as the value
            interval_sets: dict
                A dictionary of interval set data, with the name as the key
                and the data as the value
            resolution_mapping: dict
                The mapping of spatial and temporal resolutions to
                inputs, outputs and scenarios
        """
        return {
            "timesteps": self.timesteps,
            "max_iterations": self.max_iterations,
            "sector_model_config": self.sector_model_data,
            "scenario_data": self.scenario_data,
            "planning": self.planning,
            "region_sets": self.regions,
            "interval_sets": self.time_intervals,
            "resolution_mapping": self.resolution_mapping
        }

    def load_sos_config(self):
        """Parse model master config

        - configures run mode
        - sets max iterations for solving interdependencies
        - points to timesteps file
        - points to shared data files
        - points to sector models and sector model data files
        """
        msg = "Looking for configuration data in {}".format(self._config_file_path)
        self.logger.info(msg)

        data = load(self._config_file_path)
        validate_sos_model_config(data)
        return data

    def load_timesteps(self):
        """Parse model timesteps
        """
        file_path = self._get_path_from_config(self._config['timesteps'])
        os.path.exists(file_path)

        data = load(file_path)
        validate_timesteps(data, file_path)

        return data

    def load_max_iterations(self):
        """Parse max_iterations setting
        """
        if "max_iterations" in self._config \
                and isinstance(self._config["max_iterations"], int) \
                and self._config["max_iterations"] > 0:
            self.max_iterations = self._config["max_iterations"]

    def load_sector_model_data(self):
        """Parse list of sector models to run

        Model details include:
        - model name
        - model config directory
        - SectorModel class name to call
        """
        return self._config['sector_models']

    def load_scenario_data(self):
        """Load scenario data from list in sos model config

        Working assumptions:

        - scenario data is list of dicts, each like::

            {
                'parameter': 'parameter_name',
                'file': 'relative file path',
                'spatial_resolution': 'national'
                'temporal_resolution': 'annual'
            }

        - data in file is list of dicts, each like::

            {
                'value': 100,
                'units': 'kg',
                # optional, depending on parameter type:
                'region': 'UK',
                'year': 2015
            }

        Returns
        -------
        dict
            A dictionary where keys are parameters names and values are the file contents,
            so a list of dicts

        """
        scenario_data = {}
        if 'scenario_data' in self._config:
            for data_type in self._config['scenario_data']:

                name = data_type['parameter']

                spatial_res = data_type['spatial_resolution']
                temporal_res = data_type['temporal_resolution']

                self.resolution_mapping['scenario'][name] = \
                    {'spatial_resolution': spatial_res,
                     'temporal_resolution': temporal_res}

                file_path = self._get_path_from_config(data_type['file'])
                self.logger.debug("Loading scenario data from %s with %s and %s",
                                  file_path, spatial_res, temporal_res)
                data = load(file_path)
                validate_scenario_data(data, file_path)

                scenario_data[name] = data

        return scenario_data

    def load_planning(self):
        """Loads the set of build instructions for planning

        Returns
        -------
        list
            A list of planning instructions loaded from the planning file
        """
        if self._config['planning']['pre_specified']['use']:
            planning_relative_paths = self._config['planning']['pre_specified']['files']
            planning_instructions = []

            for rel_path in planning_relative_paths:
                file_path = self._get_path_from_config(rel_path)
                if os.path.exists(file_path):
                    data = load(file_path)
                    if isinstance(data, list):
                        planning_instructions.extend(data)
                    else:
                        self.logger.warning("Invalid planning instructions in %s", file_path)
                else:
                    self.logger.warning("Planning file %s not found", file_path)

            return planning_instructions
        else:
            return []

    def _get_path_from_config(self, path):
        """Return an absolute path, given a path provided from a config file

        If the provided path is relative, join it to the config file directory
        """
        if os.path.isabs(path):
            return os.path.normpath(path)
        else:
            return os.path.normpath(os.path.join(self._config_file_dir, path))

    def load_time_intervals(self):
        """Within-year time intervals are specified in ``data/<sectormodel>/time_intervals.yaml``

        These specify the mapping of model timesteps to durations within a year
        (assume modelling 365 days: no extra day in leap years, no leap seconds)

        Each time interval must have
        - start (period since beginning of year)
        - end (period since beginning of year)
        - id (label to use when passing between integration layer and sector model)

        use ISO 8601[1]_ duration format to specify periods::

            P[n]Y[n]M[n]DT[n]H[n]M[n]S

        For example::

            P1Y == 1 year
            P3M == 3 months
            PT168H == 168 hours

        So to specify a period from the beginning of March to the end of May::

            start: P2M
            end: P5M
            id: spring

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/ISO_8601#Durations

        """
        time_interval_data = {}
        if 'interval_sets' in self._config:
            for interval_set in self._config['interval_sets']:
                file_path = self._get_path_from_config(interval_set['file'])
                self.logger.debug("Loading time interval data from %s", file_path)
                data = load(file_path)
                validate_time_intervals(data, file_path)
                time_interval_data[interval_set["name"]] = data

        return time_interval_data

    def load_regions(self):
        """Model regions are specified in ``data/<sectormodel>/regions.*``

        The file format must be possible to parse with GDAL, and must contain
        an attribute "name" to use as an identifier for the region.
        """
        region_set_data = {}
        assert isinstance(self._config, dict)
        if 'region_sets' in self._config:
            for region_set in self._config['region_sets']:
                file_path = self._get_path_from_config(region_set['file'])
                self.logger.debug("Loading region set data from %s", file_path)
                data = self._parse_region_data(file_path)
                region_set_data[region_set["name"]] = data

        return region_set_data

    def _parse_region_data(self, path):
        """
        Read regions from GDAL-readable files

        Arguments
        ---------
        path: str
            Path to regions file

        Returns
        -------
        data: list
            A list of Fiona feature collections

        """
        with fiona.drivers():
            with fiona.open(path) as src:
                data = [f for f in src]

        return data



"""Handles conversion between the sets of regions used in the `SosModel`
"""
import logging
from collections import defaultdict, namedtuple

from rtree import index
from shapely.geometry import shape

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


def proportion_of_a_intersecting_b(shape_a, shape_b):
    """Calculate the proportion of shape a that intersects with shape b
    """
    intersection = shape_a.intersection(shape_b)
    return intersection.area / shape_a.area


NamedShape = namedtuple('NamedShape', ['name', 'shape'])


class RegionSet(object):
    """Hold a set of regions, spatially indexed for ease of lookup when
    constructing conversion matrices.

    Parameters
    ----------
    set_name : str
        Name to use as identifier for this set of regions
    fiona_shape_iter: iterable
        Iterable (probably a list or a reader handle)
        of fiona feature records e.g. the 'features' entry of
        a GeoJSON collection

    """
    def __init__(self, set_name, fiona_shape_iter):
        self.name = set_name
        self._regions = [
            NamedShape(region['properties']['name'], shape(region['geometry']))
            for region in fiona_shape_iter
        ]

        self._idx = index.Index()
        for pos, region in enumerate(self._regions):
            self._idx.insert(pos, region.shape.bounds)

    def intersection(self, bounds):
        """Return the subset of regions intersecting with a bounding box
        """
        return [self._regions[pos] for pos in self._idx.intersection(bounds)]

    def __getitem__(self, key):
        return self._regions[key]

    def __len__(self):
        return len(self._regions)


class RegionRegister(object):
    """Holds the sets of regions used by the SectorModels and provides conversion
    between data values relating to compatible sets of regions.
    """
    def __init__(self):
        self._register = {}
        self._conversions = defaultdict(dict)
        self.logger = logging.getLogger(__name__)

    @property
    def region_set_names(self):
        """Names of registered region sets

        Returns
        -------
        sets: list of str
        """
        return list(self._register.keys())

    def get_regions_in_set(self, set_name):
        """Return regions for a given set
        """
        if set_name in self._register:
            return self._register[set_name]
        else:
            raise ValueError("Region set {} not registered".format(set_name))

    def register(self, region_set):
        """Register a set of regions as a source/target for conversion
        """
        already_registered = self.region_set_names
        self._register[region_set.name] = region_set
        for other_set_name in already_registered:
            self._generate_coefficients(region_set, self._register[other_set_name])

    def convert(self, data, from_set_name, to_set_name):
        """Convert a list of data points for a given set of regions
        to another set of regions.

        Parameters
        ----------
        data: dict
        from_set_name: str
        to_set_name: str

        """
        converted = defaultdict(float)
        for from_key, from_value in data.items():
            coefficents = self._conversions[from_set_name][to_set_name][from_key]
            for to_key, coef in coefficents:
                converted[to_key] += coef*from_value
        return converted

    def _generate_coefficients(self, set_a, set_b):
        # from a to b
        self._conversions[set_a.name][set_b.name] = self._conversion_coefficients(set_a, set_b)
        # from b to a
        self._conversions[set_b.name][set_a.name] = self._conversion_coefficients(set_b, set_a)

    @staticmethod
    def _conversion_coefficients(from_set, to_set):
        """Return a dict containing the proportions of to_regions intersecting
        with each given from_region::

            {
                to_region.name: [
                    (from_region.name, proportion),
                    # ...
                ],
                # ...
            }

        """
        coefficients = defaultdict(list)

        for to_region in to_set:
            intersecting_from_regions = from_set.intersection(to_region.shape.bounds)

            for from_region in intersecting_from_regions:
                proportion = proportion_of_a_intersecting_b(from_region.shape, to_region.shape)
                coefficient_pair = (to_region.name, proportion)
                coefficients[from_region.name].append(coefficient_pair)

        return coefficients



"""Handles conversion between the set of time intervals used in the `SosModel`

There are three main classes, which are currently rather intertwined.
:class:`Interval` represents an individual definition of a period
within a year.
This is specified using the ISO8601 period syntax and exposes
methods which use the isodate library to parse this into an internal hourly
representation of the period.

:class:`TimeIntervalRegister` holds the definitions of time-interval sets
specified for the sector models at the :class:`~smif.sos_model.SosModel`
level.
This class exposes one public method,
:py:meth:`~TimeIntervalRegister.add_interval_set` which allows the SosModel
to add an interval definition from a model configuration to the register.

:class:`TimeSeries` is used to encapsulate any data associated with a
time interval definition set, and handles conversion from the current time
interval resolution to a target time interval definition held in the register.

Quantities
----------
Quantities are associated with a duration, period or interval.
For example 120 GWh of electricity generated during each week of February.::

        Week 1: 120 GW
        Week 2: 120 GW
        Week 3: 120 GW
        Week 4: 120 GW

Other examples of quantities:

- greenhouse gas emissions
- demands for infrastructure services
- materials use
- counts of cars past a junction
- costs of investments, operation and maintenance

Upscale: Divide
~~~~~~~~~~~~~~~

To convert to a higher temporal resolution, the values need to be apportioned
across the new time scale. In the above example, the 120 GWh of electricity
would be divided over the days of February to produce a daily time series
of generation.  For example::

        1st Feb: 17 GWh
        2nd Feb: 17 GWh
        3rd Feb: 17 GWh
        ...

Downscale: Sum
~~~~~~~~~~~~~~

To resample weekly values to a lower temporal resolution, the values
would need to be accumulated.  A monthly total would be::

        Feb: 480 GWh

Remapping
---------

Remapping quantities, as is required in the conversion from energy
demand (hourly values over a year) to energy supply (hourly values
for one week for each of four seasons) requires additional
averaging operations.  The quantities are averaged over the
many-to-one relationship of hours to time-slices, so that the
seasonal-hourly timeslices in the model approximate the hourly
profiles found across the particular seasons in the year. For example::

        hour 1: 20 GWh
        hour 2: 15 GWh
        hour 3: 10 GWh
        ...
        hour 8592: 16 GWh
        hour 8593: 12 GWh
        hour 8594: 21 GWh
        ...
        hour 8760: 43 GWh

To::

        season 1 hour 1: 20+16+.../4 GWh # Denominator number hours in sample
        season 1 hour 2: 15+12+.../4 GWh
        season 1 hour 3: 10+21+.../4 GWh
        ...

Prices
------

Unlike quantities, prices are associated with a point in time.
For example a spot price of £870/GWh.  An average price
can be associated with a duration, but even then,
we are just assigning a price to any point in time within a
range of times.

Upscale: Fill
~~~~~~~~~~~~~

Given a timeseries of monthly spot prices, converting these
to a daily price can be done by a fill operation.
E.g. copying the monthly price to each day.

From::

        Feb: £870/GWh

To::

        1st Feb: £870/GWh
        2nd Feb: £870/GWh
        ...

Downscale: Average
~~~~~~~~~~~~~~~~~~

On the other hand, going down scale, such as from daily prices
to a monthly price requires use of an averaging function. From::

        1st Feb: £870/GWh
        2nd Feb: £870/GWh
        ...

To::

        Feb: £870/GWh

Development Notes
-----------------

- We could use :py:meth:`numpy.convolve` to compare time intervals as hourly arrays
  before adding them to the set of intervals

"""
import logging
from collections import OrderedDict
from datetime import datetime, timedelta

import numpy as np
from isodate import parse_duration

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


"""Used as the reference year for computing time intervals
"""
BASE_YEAR = 2010


class Interval(object):
    """A time interval

    Parameters
    ----------
    id: str
        The unique name of the Interval
    list_of_intervals: str
        A list of tuples of valid ISO8601 duration definition
        string denoting the time elapsed from the beginning
        of the year to the (beginning, end) of the interval
    base_year: int, default=2010
        The reference year used for conversion to a datetime tuple

    Example
    -------

            >>> a = Interval('id', ('PT0H', 'PT1H'))
            >>> a.interval = ('PT1H', 'PT2H')
            >>> repr(a)
            "Interval('id', [('PT0H', 'PT1H'), ('PT1H', 'PT2H')], base_year=2010)"
            >>> str(a)
            "Interval 'id' starts at hour 0 and ends at hour 1"

    """

    def __init__(self, name, list_of_intervals, base_year=BASE_YEAR):
        self._name = name
        self._baseyear = base_year

        if len(list_of_intervals) == 0:
            msg = "Must construct Interval with at least one interval"
            raise ValueError(msg)

        if isinstance(list_of_intervals, list):
            for interval in list_of_intervals:
                assert isinstance(interval, tuple), "Interval must be constructed with tuples"
                if len(interval) != 2:
                    msg = "Interval tuple must take form (<start>, <end>)"
                    raise ValueError(msg)
            self._interval = list_of_intervals
        elif isinstance(list_of_intervals, tuple):
            self._interval = []
            self._interval.append(list_of_intervals)
        else:
            msg = "Interval tuple must take form (<start>, <end>)"
            raise ValueError(msg)

        self._validate()

    def _validate(self):
        for lower, upper in self.to_hours():
            if lower > upper:
                msg = "A time interval must not end before it starts - found %d > %d"
                raise ValueError(msg, lower, upper)

    @property
    def name(self):
        return self._name

    @property
    def start(self):
        """The start hour of the interval(s)

        Returns
        -------
        list
            A list of integers, representing the hour from the beginning of the
            year associated with the start of each of the intervals
        """
        if len(self._interval) == 1:
            return self._interval[0][0]
        else:
            return [x[0] for x in self._interval]

    @property
    def end(self):
        """The end hour of the interval(s)

        Returns
        -------
        list
            A list of integers, representing the hour from the beginning of the
            year associated with the end of each of the intervals
        """
        if len(self._interval) == 1:
            return self._interval[0][1]
        else:
            return [x[1] for x in self._interval]

    @property
    def interval(self):
        """The list of intervals

        Setter appends a tuple or list of intervals to the
        list of intervals
        """
        return sorted(self._interval)

    @interval.setter
    def interval(self, value):
        if isinstance(value, tuple):
            self._interval.append(value)
        elif isinstance(value, list):
            for element in value:
                assert isinstance(element, tuple), "A time interval must be a tuple"
            self._interval.extend(value)
        else:
            msg = "A time interval must add either a single tuple or a list of tuples"
            raise ValueError(msg)

        self._validate()

    @property
    def baseyear(self):
        """The reference year
        """
        return self._baseyear

    def __repr__(self):
        msg = "Interval('{}', {}, base_year={})"
        return msg.format(self.name, self.interval, self.baseyear)

    def __str__(self):
        string = "Interval '{}' maps to:\n".format(self.name)
        for interval in self.to_hours():
            start = interval[0]
            end = interval[1]
            suffix = "  hour {} to hour {}\n".format(start, end)
            string += suffix

        return string

    def __eq__(self, other):
        if (self.name == other.name) \
           and (self.interval == other.interval) \
           and (self.baseyear == other.baseyear):
            return True
        else:
            return False

    def to_hours(self):
        """Return a list of tuples of the intervals in terms of hours

        Returns
        -------
        list
            A list of tuples of the start and end hours of the year
            of the interval

        """
        hours = []
        for start_interval, end_interval in self.interval:
            start = self._convert_to_hours(start_interval)
            end = self._convert_to_hours(end_interval)
            hours.append((start, end))
        return hours

    def _convert_to_hours(self, duration):
        """

        Parameters
        ----------
        duration: str
            A valid ISO8601 duration definition string

        Returns
        -------
        int
            The hour in the year associated with the duration

        """
        reference = datetime(self._baseyear, 1, 1, 0)
        parsed_duration = parse_duration(duration)
        if isinstance(parsed_duration, timedelta):
            hours = parsed_duration.days * 24 + \
                    parsed_duration.seconds // 3600
        else:
            time = parsed_duration.totimedelta(reference)
            hours = time.days * 24 + time.seconds // 3600
        return hours

    def to_hourly_array(self):
        """Converts a list of intervals to a boolean array of hours

        """
        array = np.zeros(8760, dtype=np.int)
        list_of_tuples = self.to_hours()
        for lower, upper in list_of_tuples:
            array[lower:upper] += 1
        return array


class TimeSeries(object):
    """A series of values associated with an interval definition

    Parameters
    ----------
    data: list
        A list of dicts, each entry containing 'name' and 'value' keys
    """
    def __init__(self, data):
        self.logger = logging.getLogger(__name__)
        values = []
        name_list = []
        for row in data:
            name_list.append(row['id'])
            values.append(row['value'])
        self.names = name_list
        self.values = values
        self._hourly_values = np.zeros(8760, dtype=np.float64)

    @property
    def hourly_values(self):
        """The timeseries resampled to hourly values
        """
        return self._hourly_values


class TimeIntervalRegister:
    """Holds the set of time-intervals used by the SectorModels

    Parameters
    ----------
    base_year: int, default=2010
        Set the year which is used as a reference by all time interval sets
        and repeated for each future year
    """

    def __init__(self, base_year=2010):
        self._base_year = base_year
        self._register = {}
        self.logger = logging.getLogger(__name__)

    @property
    def interval_set_names(self):
        """A list of the interval set names contained in the register

        Returns
        -------
        list
        """
        return list(self._register.keys())

    def get_intervals_in_set(self, set_name):
        """

        Parameters
        ----------
        set_name: str
            The unique identifying name of the interval definitions

        Returns
        -------
        :class:`collections.OrderedDict`
            Returns a collection of the intervals in the order in which they
            were defined

        """
        self._check_interval_in_register(set_name)
        return self._register[set_name]

    def register(self, intervals, set_name):
        """Add a time-interval definition to the set of intervals types

        Detects duplicate references to the same annual-hours by performing a
        convolution of the two one-dimensional arrays of time-intervals.

        Parameters
        ----------
        intervals: list
            Time intervals required as a list of dicts, with required keys
            ``start``, ``end`` and ``name``
        set_name: str
            A unique identifier for the set of time intervals

        """
        if set_name in self._register:
            msg = "An interval set named {} has already been loaded"
            raise ValueError(msg.format(set_name))

        self._register[set_name] = OrderedDict()

        for interval in intervals:

            name = interval['id']
            self.logger.debug("Adding interval '%s' to set '%s'", name, set_name)

            if name in self._register[set_name]:
                interval_tuple = (interval['start'], interval['end'])
                self._register[set_name][name].interval = interval_tuple
            else:
                interval_tuple = (interval['start'], interval['end'])
                self._register[set_name][name] = Interval(name,
                                                          interval_tuple,
                                                          self._base_year)

        self.logger.info("Adding interval set '%s' to register", set_name)
        self._validate_intervals()

    def _check_interval_in_register(self, interval):
        """

        Parameters
        ----------
        interval: str
            The name of the interval to look for

        """
        if interval not in self._register:
            msg = "The interval set '{}' is not in the register"
            raise ValueError(msg.format(interval))

    def convert(self, timeseries, from_interval, to_interval):
        """Convert some data to a time_interval type

        Parameters
        ----------
        timeseries: :class:`~smif.convert.interval.TimeSeries`
            The timeseries to convert from `from_interval` to `to_interval`
        from_interval: str
            The unique identifier of a interval type which matches the
            timeseries
        to_interval: str
            The unique identifier of a registered interval type

        Returns
        -------
        dict
            A dictionary with keys `name` and `value`, where the entries
            for `key` are the name of the target time interval, and the
            values are the resampled timeseries values.

        """
        results = []

        self._check_interval_in_register(from_interval)
        self._convert_to_hourly_buckets(timeseries, from_interval)

        target_intervals = self.get_intervals_in_set(to_interval)
        for name, interval in target_intervals.items():
            self.logger.debug("Resampling to %s", name)
            interval_tuples = interval.to_hours()

            total = 0

            for lower, upper in interval_tuples:
                self.logger.debug("Range: %s-%s", lower, upper)
                total += sum(timeseries.hourly_values[lower:upper])

            results.append({'id': name,
                            'value': total})

        return results

    def _convert_to_hourly_buckets(self, timeseries, interval_set):
        """Iterates through the `timeseries` and assigns values to hourly buckets

        Parameters
        ----------
        timeseries: :class:`~smif.convert.interval.TimeSeries`
            The timeseries to convert to hourly buckets ready for further
            operations
        interval_set: str
            The name of the interval set associated with the timeseries

        """
        for name, value in zip(timeseries.names, timeseries.values):
            list_of_intervals = self._register[interval_set][name].to_hours()
            divisor = len(list_of_intervals)
            for lower, upper in list_of_intervals:
                self.logger.debug("lower: %s, upper: %s", lower, upper)
                number_hours_in_range = upper - lower
                self.logger.debug("number_hours: %s", number_hours_in_range)

                apportioned_value = float(value) / number_hours_in_range
                self.logger.debug("apportioned_value: %s", apportioned_value)
                timeseries.hourly_values[lower:upper] = apportioned_value / divisor

    def _get_hourly_array(self, set_name):
        """
        """
        array = np.zeros(8760, dtype=np.int)
        for interval in self.get_intervals_in_set(set_name).values():
            array += interval.to_hourly_array()
        return array

    def _validate_intervals(self):
        for set_name in self._register.keys():
            array = self._get_hourly_array(set_name)
            duplicate_hours = np.where(array > 1)[0]
            if len(duplicate_hours) == 0:
                self.logger.debug("No duplicate hours in %s", set_name)
            else:
                hour = duplicate_hours[0]
                msg = "Duplicate entry for hour {} in interval set {}."
                self.logger.warning(msg.format(hour, set_name))
                raise ValueError(msg.format(hour, set_name))



# -*- coding: utf-8 -*-
import os

from smif.data_layer.sos_model_config import SosModelReader
from smif.data_layer.validate import VALIDATION_ERRORS


class TestSosModelReader():

    def _get_model_config(self, folder):
        return os.path.join(str(folder), "config", "model.yaml")

    def _get_reader(self, folder):
        return SosModelReader(self._get_model_config(folder))

    def test_read_sos_model(self, setup_project_folder):

        reader = self._get_reader(setup_project_folder)
        reader.load()

        # check timesteps filename
        expected = 'timesteps.yaml'
        assert reader._config['timesteps'] == expected

        # check planning filename list
        expected = ['../data/water_supply/pre-specified.yaml']
        assert reader._config['planning']['pre_specified']['files'] == expected

    def test_model_list(self, setup_project_folder):

        reader = self._get_reader(setup_project_folder)
        reader.load()

        expected = [
            {
                "name": "water_supply",
                "path": "../models/water_supply/__init__.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                "initial_conditions": [
                    "../data/water_supply/initial_conditions/assets_1.yaml"
                ],
                "interventions": [
                    "../data/water_supply/interventions/water_asset_abc.yaml"
                ]
            }
        ]

        assert reader.sector_model_data == expected
        assert reader.data["sector_model_config"] == expected

    def test_timesteps(self, setup_project_folder):

        reader = self._get_reader(setup_project_folder)
        reader.load()

        expected = [2010, 2011, 2012]

        assert reader.timesteps == expected
        assert reader.data["timesteps"] == expected

    def test_abs_path_for_timesteps(self, setup_project_folder, setup_abs_path_to_timesteps):
        """Expect absolute paths to work fine
        """
        reader = self._get_reader(setup_project_folder)
        reader.load()

        expected = [2010, 2011, 2012]

        assert reader.timesteps == expected
        assert reader.data["timesteps"] == expected

    def test_timesteps_alternate_file(self, setup_project_folder,
                                      setup_config_file_timesteps_two,
                                      setup_timesteps_file_two):

        reader = self._get_reader(setup_project_folder)
        reader.load()

        expected = [2015, 2020, 2025]
        actual = reader.timesteps
        assert actual == expected

    def test_timesteps_invalid(self, setup_project_folder, setup_timesteps_file_invalid):
        """Expect an error on trying to load an invalid file
        """
        reader = self._get_reader(setup_project_folder)

        reader.load()
        ex = VALIDATION_ERRORS.pop()
        assert "expected a list of timesteps" in str(ex)

    def test_load_scenario_data(self, setup_project_folder, setup_scenario_data):
        """Load a population parameter from scenario config files
        """
        reader = self._get_reader(setup_project_folder)
        reader.load()
        data = reader.data["scenario_data"]
        assert "population" in data
        assert len(data["population"]) == 3
        assert data["population"][0] == {
            'value': 100,
            'units': 'people',
            'region': 'GB',
            'year': 2015
        }

    def test_load_scenario_data_missing(self):
        """Expect empty dict as default if scenario data missing
        """
        reader = SosModelReader('test.yaml')
        reader.load_scenario_data()
        assert reader.scenario_data == {}

    def test_no_planning(self, setup_project_folder, setup_no_planning):
        """Expect an empty list if planning is not to be used
        """
        reader = self._get_reader(setup_project_folder)
        reader.load()
        assert reader.data["planning"] == []

    def test_planning_missing(self, setup_project_folder, setup_planning_missing):
        """Expect an empty list and warning if planning file is missing
        """
        reader = self._get_reader(setup_project_folder)
        reader.load()
        assert reader.data["planning"] == []

    def test_planning_empty(self, setup_project_folder, setup_planning_empty):
        """Expect an empty list and warning if planning file is empty
        """
        reader = self._get_reader(setup_project_folder)
        reader.load()
        assert reader.data["planning"] == []

    def test_load_regions_shapefile(self, setup_project_folder):
        reader = self._get_reader(setup_project_folder)
        reader.load()
        data = reader.data['region_sets']
        assert len(data) == 1



"""Test command line interface
"""

import os
import subprocess
from tempfile import TemporaryDirectory
from unittest.mock import call, patch

import smif
from pytest import raises
from smif.cli import (confirm, parse_arguments, setup_project_folder,
                      validate_config)
from smif.data_layer.validate import VALIDATION_ERRORS


def get_args(args):
    """Get args object from list of strings
    """
    parser = parse_arguments()
    return parser.parse_args(args)


def test_parse_arguments():
    """Setup a project folder argument parsing
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', project_folder])

        expected = project_folder
        actual = args.path
        assert actual == expected

        # Ensure that the `setup_configuration` function is called when `setup`
        # command is passed to the cli
        assert args.func.__name__ == 'setup_configuration'


def test_fixture_single_run_valid():
    """Test validating the filesystem-based single_run fixture
    """
    config_file = os.path.join(os.path.dirname(__file__),
                               '..', 'fixtures', 'single_run', 'config', 'model.yaml')
    output = subprocess.run(["smif", "validate", config_file], stdout=subprocess.PIPE)
    assert "The model configuration was valid" in str(output.stdout)


def test_fixture_single_run():
    """Test running the filesystem-based single_run fixture
    """
    config_file = os.path.join(os.path.dirname(__file__),
                               '..', 'fixtures', 'single_run', 'config', 'model.yaml')
    output = subprocess.run(["smif", "run", config_file], stdout=subprocess.PIPE)
    assert "Model run complete" in str(output.stdout)


def test_setup_project_folder():
    """Test contents of the setup project folder
    """
    with TemporaryDirectory() as project_folder:
        setup_project_folder(project_folder)

        assert os.path.exists(project_folder)

        folder_list = ['data']
        for folder in folder_list:
            folder_path = os.path.join(project_folder, folder)
            print(folder_path)
            assert os.path.exists(folder_path)


def test_run_sector_model(setup_folder_structure):
    """Run a sector model in the list
    """
    config_file = os.path.join(str(setup_folder_structure), 'config', 'model.yaml')
    args = get_args(['run', '--model', 'water_supply', config_file])

    expected = 'water_supply'
    actual = args.model
    assert actual == expected


def test_dont_run_invalid_sector_model(setup_folder_structure, setup_project_folder):
    """Don't try to run a sector model which is not in the list
    """
    model_name = 'invalid_model_name'
    config_file = os.path.join(str(setup_folder_structure), 'config', 'model.yaml')
    args = get_args(['run', '-m', model_name, config_file])

    assert args.model == model_name
    assert args.path == config_file


def test_validation_call(setup_folder_structure, setup_project_folder):
    """Ensure validation gets called
    """
    config_file = os.path.join(str(setup_folder_structure), 'config', 'model.yaml')
    args = get_args(['validate', config_file])

    expected = config_file
    actual = args.path
    assert actual == expected
    assert args.func.__name__ == 'validate_config'


@patch('smif.cli.LOGGER.error')
def test_validation_no_file(error_logger):
    """Expect error and quit if model config is missing
    """
    args = get_args(['validate', '/path/to/missing_file.yaml'])

    with raises(SystemExit):
        validate_config(args)

    path = os.path.abspath('/path/to/missing_file.yaml')
    msg = "The model configuration file '%s' was not found"
    error_logger.assert_called_with(msg, path)


@patch('builtins.print')
def test_validation_valid(mock_print, setup_project_folder):
    """Ensure configuration file is valid
    """
    config_file = os.path.join(str(setup_project_folder), 'config', 'model.yaml')
    args = get_args(['validate', config_file])

    validate_config(args)
    mock_print.assert_called_with('The model configuration was valid')


@patch('smif.cli.LOGGER.error')
@patch('builtins.print')
def test_validation_invalid(
        mock_print,
        error_logger,
        setup_folder_structure,
        setup_project_folder,
        setup_timesteps_file_invalid):
    """Ensure invalid configuration file raises error
    """
    config_file = os.path.join(str(setup_folder_structure), 'config', 'model.yaml')
    args = get_args(['validate', config_file])

    with raises(SystemExit):
        validate_config(args)

    assert len(VALIDATION_ERRORS) > 0
    assert error_logger.called
    mock_print.assert_called_with('The model configuration was invalid')


@patch('builtins.input', return_value='y')
def test_confirm_yes(input):
    assert confirm()


@patch('builtins.input', return_value='n')
def test_confirm_no(input):
    assert not confirm()


@patch('builtins.input', return_value='')
def test_confirm_default_response(input):
    assert not confirm()


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_default_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: ')])


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_custom_message(mock_print, input):
    confirm('Create directory?', True)
    input.assert_has_calls([call('Create directory? [y]|n: ')])


@patch('builtins.input', side_effect=['invalid', 'y'])
@patch('builtins.print')
def test_confirm_repeat_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: '), call('Confirm [n]|y: ')])
    mock_print.assert_called_with('please enter y or n.')


def test_version_display():
    """Expect version number from `smif -V`
    """
    output = subprocess.run(['smif', '-V'], stdout=subprocess.PIPE)
    assert smif.__version__ in str(output.stdout)


def test_verbose_debug():
    """Expect debug message from `smif -vv`
    """
    output = subprocess.run(['smif', '-vv'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_debug_alt():
    """Expect debug message from `smif --verbose --verbose`
    """
    output = subprocess.run(['smif', '--verbose', '--verbose'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_info(setup_folder_structure, setup_project_folder):
    """Expect info message from `smif -v validate <config_file>`
    """
    config_file = os.path.join(str(setup_folder_structure), 'config', 'model.yaml')
    output = subprocess.run(['smif', '-v', 'validate', config_file], stderr=subprocess.PIPE)
    assert 'INFO' in str(output.stderr)



"""Tests the definition and solution of the planning problem optimisation
"""

from smif.intervention import Intervention, InterventionRegister
from smif.optimisation import (feature_vfa_model, formulate_model,
                               linear_vfa_model, solve_model, state_vfa_model)


def test_linear_vfa_model():

    assets = ['asset_one', 'asset_two']
    constraint = {'asset_one': 1, 'asset_two': 1}
    costs = {'asset_one': 1200, 'asset_two': 1000}
    value = {'asset_one': 1300, 'asset_two': 100}

    model = linear_vfa_model(assets, constraint, costs, value)
    results = solve_model(model)

    assert results.x['asset_one'].value == 0.0
    assert results.x['asset_two'].value == 0.0


def test_state_vfa_model():
    """Do everything, when that's cheaper than doing nothing

    States:

    #   a  b
    1 - 0  0
    2 - 0  1
    3 - 1  0
    4 - 1  1

    """

    assets = ['asset_one', 'asset_two']
    constraint = {'asset_one': 1, 'asset_two': 1}
    costs = {'asset_one': 1200, 'asset_two': 1000}
    states = {('asset_one', 1): 0,
              ('asset_one', 2): 0,
              ('asset_one', 3): 1,
              ('asset_one', 4): 1,
              ('asset_two', 1): 0,
              ('asset_two', 2): 1,
              ('asset_two', 3): 0,
              ('asset_two', 4): 1}

    value = {1: 2200, 2: 1200, 3: 1000, 4: -1}

    model = state_vfa_model(assets, constraint, costs, value, states)
    results = solve_model(model)

    assert results.x['asset_one'].value == 1.0
    assert results.x['asset_two'].value == 1.0


def test_state_vfa_model_one():
    """Take no action when inaction is cheaper than action

    """

    assets = ['asset_one', 'asset_two']
    constraint = {'asset_one': 1, 'asset_two': 1}
    costs = {'asset_one': 1200, 'asset_two': 1000}
    states = {('asset_one', 1): 0,
              ('asset_one', 2): 0,
              ('asset_one', 3): 1,
              ('asset_one', 4): 1,
              ('asset_two', 1): 0,
              ('asset_two', 2): 1,
              ('asset_two', 3): 0,
              ('asset_two', 4): 1}

    value = {1: 0, 2: 0, 3: 0, 4: 0}

    model = state_vfa_model(assets, constraint, costs, value, states)
    results = solve_model(model)

    assert results.x['asset_one'].value == 0.0
    assert results.x['asset_two'].value == 0.0


def test_features_vfa_model():
    """Use basis functions as the value function approximation

    Passes in a state as a list of asset names

    """
    assets = ['asset_one', 'asset_two']
    constraint = {'asset_one': 1, 'asset_two': 1}
    costs = {'asset_one': 40000, 'asset_two': 500000}

    feature_coefficients = {'bigness': 10,
                            'colourfulness': 200,
                            'powerfulness': 3000}
    asset_features = {('bigness', 'asset_one'): 1,
                      ('colourfulness', 'asset_one'): 1,
                      ('powerfulness', 'asset_one'): 0,
                      ('bigness', 'asset_two'): 0,
                      ('colourfulness', 'asset_two'): 1,
                      ('powerfulness', 'asset_two'): 1}

    state = ['asset_one']
    model = feature_vfa_model(assets, constraint, costs,
                              feature_coefficients, asset_features)
    results = solve_model(model, state)
    print(results)
    assert results.x['asset_one'].value == 1.0
    assert results.x['asset_two'].value == 0.0
    assert results.OBJ() == 40210

    state = ['asset_two']
    model = feature_vfa_model(assets, constraint, costs,
                              feature_coefficients, asset_features)
    results = solve_model(model, state)
    print(results)
    assert results.x['asset_one'].value == 0.0
    assert results.x['asset_two'].value == 1.0
    assert results.OBJ() == 503200


def get_asset_one():
    name = 'asset_one'
    data = {
        'sector': 'water_supply',
        'capacity': {
            'units': 'ML/day',
            'value': 5},
        'capital cost': {
            'units': '£',
            'value': 10e9},
        'location': 'oxford'
        }
    return Intervention(name=name, data=data)


def get_asset_two():
    name = 'asset_two'
    data = {
        'sector': 'water_supply',
        'capacity': {
            'units': 'ML/day',
            'value': 4.5},
        'capital cost': {
            'units': '£',
            'value': 12e9},
        'location': 'oxford'
        }
    return Intervention(name=name, data=data)


def test_passing_asset_register_to_model():
    """Tests passing a `smif.asset.InterventionRegister` to formulate the model
    """
    asset_one = get_asset_one()
    asset_two = get_asset_two()
    register = InterventionRegister()
    register.register(asset_one)
    register.register(asset_two)

    constraint = {'asset_one': 1, 'asset_two': 1}

    feature_coefficients = {'bigness': 10,
                            'colourfulness': 200,
                            'powerfulness': 3000}
    asset_features = {('bigness', 'asset_one'): 1,
                      ('colourfulness', 'asset_one'): 1,
                      ('powerfulness', 'asset_one'): 0,
                      ('bigness', 'asset_two'): 0,
                      ('colourfulness', 'asset_two'): 1,
                      ('powerfulness', 'asset_two'): 1}

    model = formulate_model(register, constraint, feature_coefficients,
                            asset_features)

    state = ['asset_one']
    results = solve_model(model, state)
    print(results)
    assert results.x['asset_one'].value == 1.0
    assert results.x['asset_two'].value == 0.0
    assert results.OBJ() == 10000000210



from pytest import fixture, raises
from smif.intervention import (Asset, AssetRegister, Intervention,
                               InterventionRegister)


@fixture(scope='function')
def get_wtp():
    name = 'water_treatment_plant'
    data = {
        'sector': 'water_supply',
        'capacity': {
            'units': 'ML/day',
            'value': 5
        },
        'location': 'oxford',
        'build_date': 2016
    }
    return Asset(
        name=name,
        data=data
    )


@fixture(scope='function')
def get_intervention():
    name = 'water_treatment_plant'
    data = {
        'sector': 'water_supply',
        'capacity': {
            'units': 'ML/day',
            'value': 5
        },
        'location': 'oxford'
    }
    return Intervention(
        name=name,
        data=data
    )


@fixture(scope='function')
def build_intervention_ws():
    data = {
        'sector': 'water_supply',
        'name': 'oxford treatment plant',
        'capacity': {
            'units': 'ML/day',
            'value': 450
            },
        'capital_cost': {
            'units': 'M£',
            'value': 500
        },
        'location': "POINT(51.1 -1.7)"
        }

    return data


@fixture(scope='function')
def build_intervention_es():
    data = {
        'sector': 'energy_supply',
        'name': 'London Array',
        'capacity': {
            'units': 'GW',
            'value': 2
            },
        'capital_cost': {
            'units': 'M£',
            'value': 2500
        },
        'location': "POINT(50.1 -1.4)"
        }

    return data


@fixture(scope='function')
def build_register_two(build_intervention_es, build_intervention_ws):
    water = Intervention(data=build_intervention_ws)
    energy = Intervention(data=build_intervention_es)

    register = InterventionRegister()
    register.register(energy)
    register.register(water)
    return register


class TestIntervention:

    def test_intervention_init_sector(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        assert actual.sector == 'water_supply'

    def test_intervention_init_build_date(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        with raises(AttributeError):
            assert actual.build_date

    def test_intervention_init_location(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        assert actual.location == "POINT(51.1 -1.7)"


class TestAsset:

    def test_create_asset(self, get_wtp):
        water_treatment_plant = get_wtp
        assert water_treatment_plant.name == 'water_treatment_plant'

    def test_create_asset_with_sector(self):
        data = {
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'build_date': 2020,
            'location': "POINT(51.1 -1.7)"
        }
        asset = Asset(data=data)
        assert asset.sector == 'water_supply'

    def test_create_asset_without_sector(self):
        data = {
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'build_date': 2020,
            'location': "POINT(51.1 -1.7)"
        }
        asset = Asset(data=data, sector='water_supply')
        assert asset.sector == 'water_supply'

    def test_create_asset_with_full_data(self, get_wtp):
        name = 'water_treatment_plant'
        data = {
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'build_date': 2020,
            'location': "POINT(51.1 -1.7)"
        }
        wtp = Asset(
            name,
            data
        )
        assert wtp.location == "POINT(51.1 -1.7)"
        assert wtp.build_date == 2020

    def test_location(self, get_wtp):
        water_treatment_plant = get_wtp
        assert water_treatment_plant.location == 'oxford'

    def test_set_location(self, get_wtp):
        water_treatment_plant = get_wtp
        water_treatment_plant.location = "POINT(51.1 -1.7)"
        assert water_treatment_plant.location == "POINT(51.1 -1.7)"

    def test_build_date(self, get_wtp):
        water_treatment_plant = get_wtp
        assert water_treatment_plant.build_date == 2016

    def test_set_build_date(self, get_wtp):
        water_treatment_plant = get_wtp
        water_treatment_plant.build_date = 2020
        assert water_treatment_plant.build_date == 2020

    def test_get_data(self, get_wtp):
        water_treatment_plant = get_wtp
        water_treatment_plant.location = "POINT(51.1 -1.7)"
        water_treatment_plant.build_date = 2020
        water_treatment_plant.data["name"] = "oxford treatment plant"

        assert water_treatment_plant.data == {
            'sector': 'water_supply',
            'name': 'oxford treatment plant',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'location': "POINT(51.1 -1.7)",
            'build_date': 2020
        }

    def test_hash(self, get_wtp):
        water_treatment_plant = get_wtp

        data_str = '{"build_date": 2016, ' + \
                   '"capacity": {"units": "ML/day", "value": 5}, ' + \
                   '"location": "oxford", ' + \
                   '"name": "water_treatment_plant", ' + \
                   '"sector": "water_supply"}'

        repr_str = 'Asset("water_treatment_plant", {' + \
                   '"build_date": 2016, ' + \
                   '"capacity": {"units": "ML/day", "value": 5}, ' + \
                   '"location": "oxford", ' + \
                   '"name": "water_treatment_plant", ' + \
                   '"sector": "water_supply"})'

        # should be able to reproduce sha1sum by doing
        # `printf "data_str..." | sha1sum` on the command line
        sha1sum = "c2c6e0860718e11bc85913ef4701d629e604c41f"
        assert str(water_treatment_plant) == data_str
        assert repr(water_treatment_plant) == repr_str
        assert water_treatment_plant.sha1sum() == sha1sum


class TestAssetRegister:

    def test_fail_add_intervention(self, get_intervention):
        not_an_asset = get_intervention
        register = AssetRegister()
        with raises(TypeError):
            register.register(not_an_asset)

    def test_register_assert_length(self, get_wtp):
        water = get_wtp
        register = AssetRegister()
        register.register(water)

        assert len(register) == 1

    def test_register_assert_keys(self, get_wtp):
        water = get_wtp
        register = AssetRegister()
        register.register(water)

        assert sorted(register._attribute_keys) == [
            "build_date",
            "capacity",
            "location",
            "name",
            "sector"
        ]


class TestInterventionRegister:

    def test_register_intervention(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        assert len(register) == 1
        assert sorted(register._attribute_keys) == [
            "capacity",
            "location",
            "name",
            "sector"
        ]

        attr_idx = register.attribute_index("name")
        possible = register._attribute_possible_values[attr_idx]
        assert possible == [None, "water_treatment_plant"]

        attr_idx = register.attribute_index("capacity")
        possible = register._attribute_possible_values[attr_idx]
        assert possible == [None, {'units': 'ML/day', 'value': 5}]

    def test_retrieve_intervention(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        # pick an asset from the list - this is what the optimiser will do
        numeric_asset = register._numeric_keys[0]

        asset = register.numeric_to_intervention(numeric_asset)
        assert asset.name == "water_treatment_plant"

        assert asset.data == {
            'name': 'water_treatment_plant',
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'location': 'oxford'
        }

    def test_retrieve_intervention_by_name(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        # pick an asset from the list - this is what the optimiser will do
        asset = register.get_intervention('water_treatment_plant')
        assert asset.name == "water_treatment_plant"

        assert asset.data == {
            'name': 'water_treatment_plant',
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'location': 'oxford'
        }

    def test_error_when_retrieve_by_name(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)
        with raises(ValueError) as excinfo:
            register.get_intervention('not_here')
        expected = "Intervention 'not_here' not found in register"
        assert str(excinfo.value) == expected

    def test_iterate_over_interventions(self, get_intervention):
        """Test __iter___ method of AssetRegister class

        """
        asset_one = get_intervention
        register = InterventionRegister()
        register.register(asset_one)

        for asset in register:
            assert asset.sha1sum() == asset_one.sha1sum()

    def test_add_duplicate_intervention(self, get_intervention):
        """Tests that only unique interventions are retained

        """
        asset_one = get_intervention
        asset_two = get_intervention
        register = InterventionRegister()
        register.register(asset_one)
        register.register(asset_two)

        assert len(register) == 1

    def test_register_len_one(self, build_intervention_ws):
        water = Intervention(data=build_intervention_ws)
        register = InterventionRegister()
        register.register(water)
        assert len(register) == 1

    def test_register_len_two(self, build_register_two):
        register = build_register_two
        assert len(register) == 2



# -*- coding: utf-8 -*-
import os
from unittest.mock import MagicMock

from smif.data_layer.sector_model_config import SectorModelReader


class TestSectorModelReader(object):
    def _model_config_dir(self, project_folder):
        return os.path.join(str(project_folder), 'data', 'water_supply')

    def _model_path(self, project_folder):
        return os.path.join(str(project_folder), 'models', 'water_supply', 'water_supply.py')

    def _reader(self, project_folder):
        model_name = 'water_supply'
        model_path = self._model_path(project_folder)
        model_classname = ''
        config_dir = self._model_config_dir(project_folder)
        return SectorModelReader({
            "model_name": model_name,
            "model_path": model_path,
            "model_classname": model_classname,
            "model_config_dir": config_dir,
            "initial_conditions": [
                os.path.join(config_dir, "initial_conditions/assets_1.yaml")
            ],
            "interventions": [
                os.path.join(config_dir, "interventions/water_asset_abc.yaml")
            ]
        })

    def test_reader_without_initial_config(self):
        reader = SectorModelReader()
        assert reader.model_name is None
        reader.model_name = "test_model"
        assert reader.model_name == "test_model"

    def test_load(self, setup_project_folder):
        reader = self._reader(setup_project_folder)
        reader.load()

        # could check actual data from conftest.py:setup_water_inputs
        assert reader.data["inputs"] is not None
        assert reader.data["outputs"] is not None

    def test_load_errors(self, setup_project_missing_model_config):
        reader = self._reader(setup_project_missing_model_config)
        reader.logger.warning = MagicMock()

        # expect a warning if loading no inputs
        inputs = reader.load_inputs()
        args, kwargs = reader.logger.warning.call_args

        assert args[0] == "No %s provided for '%s' model: %s not found"
        assert args[1] == "inputs"
        assert args[2] == "water_supply"
        assert "inputs.yaml" in args[3]
        assert inputs == []

        outputs = reader.load_outputs()
        args, kwargs = reader.logger.warning.call_args

        assert args[0] == "No %s provided for '%s' model: %s not found"
        assert args[1] == "outputs"
        assert args[2] == "water_supply"
        assert "outputs.yaml" in args[3]
        assert outputs == []

    def test_load_empty(self, setup_project_empty_model_io):
        reader = self._reader(setup_project_empty_model_io)
        reader.logger.warning = MagicMock()

        # expect a warning if loading no inputs
        inputs = reader.load_inputs()
        args, kwargs = reader.logger.warning.call_args

        assert args[0] == "No %s provided for '%s' model: %s was empty"
        assert args[1] == "inputs"
        assert args[2] == "water_supply"
        assert "inputs.yaml" in args[3]
        assert inputs == []

        outputs = reader.load_outputs()
        args, kwargs = reader.logger.warning.call_args

        assert args[0] == "No %s provided for '%s' model: %s was empty"
        assert args[1] == "outputs"
        assert args[2] == "water_supply"
        assert "outputs.yaml" in args[3]
        assert outputs == []

    def test_load_interventions(self, setup_project_folder):

        reader = self._reader(setup_project_folder)
        reader.load()

        expected = ['water_asset_a', 'water_asset_b', 'water_asset_c']
        actual = [asset['name'] for asset in reader.interventions]
        assert actual == expected

    def test_load_interventions_two_files(self, setup_project_folder,
                                          setup_config_file_two,
                                          setup_water_intervention_d):

        model_name = 'water_supply'
        model_path = self._model_path(setup_project_folder)
        model_classname = ''
        config_dir = self._model_config_dir(setup_project_folder)
        reader = SectorModelReader({
            "model_name": model_name,
            "model_path": model_path,
            "model_classname": model_classname,
            "model_config_dir": config_dir,
            "initial_conditions": [
                os.path.join(config_dir, "initial_conditions/assets_1.yaml")
            ],
            "interventions": [
                os.path.join(config_dir, "interventions/water_asset_abc.yaml"),
                os.path.join(config_dir, "interventions/water_asset_d.yaml")
            ]
        })

        reader.load()

        expected = ['water_asset_a', 'water_asset_b',
                    'water_asset_c', 'water_asset_d']
        actual = [asset['name'] for asset in reader.interventions]
        assert actual == expected



# -*- coding: utf-8 -*-
"""Encapsulates the input or output parameters of a sector model,
for example::

        - name: petrol_price
          spatial_resolution: GB
          temporal_resolution: annual
        - name: diesel_price
          spatial_resolution: GB
          temporal_resolution: annual
        - name: LPG_price
          spatial_resolution: GB
          temporal_resolution: annual

"""
from __future__ import absolute_import, division, print_function

import logging
from collections import namedtuple

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell, University of Oxford 2017"
__license__ = "mit"


Parameter = namedtuple(
    "Parameter",
    [
        "name",
        "spatial_resolution",
        "temporal_resolution"
    ]
)


class ModelParameters(object):
    """A container for all the model inputs

    Arguments
    =========
    inputs : list
        A list of dicts of model parameter name, spatial resolution
        and temporal resolution
    """
    def __init__(self, parameters):
        self._parameters = [Parameter(parameter['name'],
                                      parameter['spatial_resolution'],
                                      parameter['temporal_resolution']
                                      ) for parameter in parameters]
        self.logger = logging.getLogger(__name__)

    @property
    def parameters(self):
        """A list of the model parameters

        Returns
        =======
        :class:`smif.parameters.ParameterList`
        """
        return self._parameters

    def __len__(self):
        return len(self.parameters)

    def get_spatial_res(self, name):
        """The spatial resolution for parameter `name`

        Arguments
        ---------
        name: str
            The name of a model parameter
        """
        for parameter in self._parameters:
            if parameter.name == name:
                spatial_resolution = parameter.spatial_resolution
                break
        else:
            raise ValueError("No output found for name '{}'".format(name))
        return spatial_resolution

    def get_temporal_res(self, name):
        """The temporal resolution for parameter `name`

        Arguments
        ---------
        name: str
            The name of a model parameter
        """
        for parameter in self._parameters:
            if parameter.name == name:
                temporal_resolution = parameter.temporal_resolution
                break
        else:
            raise ValueError("No output found for name '{}'".format(name))
        return temporal_resolution

    @property
    def spatial_resolutions(self):
        """A list of the spatial resolutions

        Returns
        -------
        list
            A list of the spatial resolutions associated with the model
            parameters
        """
        return [parameter.spatial_resolution for parameter in self._parameters]

    @property
    def temporal_resolutions(self):
        """A list of the temporal resolutions

        Returns
        -------
        list
            A list of the temporal resolutions associated with the model
            parameters
        """
        return [parameter.temporal_resolution for parameter in self._parameters]

    @property
    def names(self):
        return [parameter.name for parameter in self._parameters]



"""Test aggregation/disaggregation of data between sets of areas
"""
from copy import copy
from pytest import fixture, raises
from shapely.geometry import shape
from smif.convert.area import (RegionRegister, RegionSet,
                               proportion_of_a_intersecting_b)


@fixture(scope='function')
def regions():
    """Return data structure for test regions/shapes
    """
    return [
        {
            'type': 'Feature',
            'properties': {'name': 'unit'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'half'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 0.5], [1, 0.5], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'two'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
            }
        }
    ]


@fixture(scope='function')
def regions_half_squares():
    """Return two adjacent square regions::

        |```|```|
        | A | B |
        |...|...|

    """
    return RegionSet('half_squares', [
        {
            'type': 'Feature',
            'properties': {'name': 'a'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'b'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
            }
        },
    ])


@fixture(scope='function')
def regions_single_half_square():
    """Return single half-size square region::

        |```|
        | A |
        |...|

    """
    return RegionSet('single_half_square', [
        {
            'type': 'Feature',
            'properties': {'name': 'a'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        }
    ])


@fixture(scope='function')
def regions_rect():
    """Return single region covering 2x1 area::

        |```````|
        |   0   |
        |.......|

    """
    return RegionSet('rect', [
        {
            'type': 'Feature',
            'properties': {'name': 'zero'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
            }
        }
    ])


@fixture(scope='function')
def regions_half_triangles():
    """Return regions split diagonally::

        |``````/|
        | 0 / 1 |
        |/......|

    """
    return RegionSet('half_triangles', [
        {
            'type': 'Feature',
            'properties': {'name': 'zero'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'one'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 2], [1, 2], [1, 0]]]
            }
        },
    ])


def test_proportion(regions):
    """Sense-check proportion calculator
    """
    unit = shape(regions[0]['geometry'])
    half = shape(regions[1]['geometry'])
    two = shape(regions[2]['geometry'])

    assert proportion_of_a_intersecting_b(unit, unit) == 1

    assert proportion_of_a_intersecting_b(unit, half) == 0.5
    assert proportion_of_a_intersecting_b(half, unit) == 1

    assert proportion_of_a_intersecting_b(unit, two) == 1
    assert proportion_of_a_intersecting_b(two, unit) == 0.5
    assert proportion_of_a_intersecting_b(half, two) == 1
    assert proportion_of_a_intersecting_b(two, half) == 0.25


class TestRegionSet():
    """Test creating, looking up, retrieving region sets
    """
    def test_create(self, regions):
        rset = RegionSet('test', regions)
        assert rset.name == 'test'
        assert len(rset) == 3
        assert rset[0].name == 'unit'
        assert rset[1].name == 'half'
        assert rset[2].name == 'two'


class TestRegionRegister():
    """Test creating registers, registering region sets, converting data
    """
    def test_create(self):
        rreg = RegionRegister()
        assert rreg.region_set_names == []

        with raises(ValueError) as ex:
            rreg.get_regions_in_set('nonexistent')
        assert "Region set nonexistent not registered" in str(ex)

    def test_convert_equal(self, regions_rect):
        rreg = RegionRegister()
        # register rect
        rreg.register(regions_rect)
        # register alt rect (same area)
        regions_rect_alt = copy(regions_rect)
        regions_rect_alt.name = 'rect_alt'
        rreg.register(regions_rect_alt)

        data = {'zero': 1}
        converted = rreg.convert(data, 'rect', 'rect_alt')
        assert data == converted

    def test_convert_to_half(self, regions_rect, regions_half_squares):
        rreg = RegionRegister()
        rreg.register(regions_rect)
        rreg.register(regions_half_squares)

        data = {'zero': 1}
        converted = rreg.convert(data, 'rect', 'half_squares')
        expected = {'a': 0.5, 'b': 0.5}
        assert converted == expected

    def test_convert_from_half(self, regions_rect, regions_half_squares):
        rreg = RegionRegister()
        rreg.register(regions_rect)
        rreg.register(regions_half_squares)

        data = {'a': 0.5, 'b': 0.5}
        converted = rreg.convert(data, 'half_squares', 'rect')
        expected = {'zero': 1}
        assert converted == expected

        data = {'a': 2, 'b': 3}
        converted = rreg.convert(data, 'half_squares', 'rect')
        expected = {'zero': 5}
        assert converted == expected

    def test_convert_to_half_not_covered(self, regions_rect, regions_single_half_square):
        rreg = RegionRegister()
        rreg.register(regions_rect)
        rreg.register(regions_single_half_square)

        data = {'zero': 3}
        converted = rreg.convert(data, 'rect', 'single_half_square')
        expected = {'a': 1.5}
        assert converted == expected

    def test_convert_from_half_not_covered(self, regions_rect, regions_single_half_square):
        rreg = RegionRegister()
        rreg.register(regions_rect)
        rreg.register(regions_single_half_square)

        data = {'a': 3}
        converted = rreg.convert(data, 'single_half_square', 'rect')
        expected = {'zero': 3}
        assert converted == expected

    def test_convert_square_to_triangle(self, regions_half_squares, regions_half_triangles):
        rreg = RegionRegister()
        rreg.register(regions_half_squares)
        rreg.register(regions_half_triangles)

        data = {'a': 1, 'b': 1}
        converted = rreg.convert(data, 'half_squares', 'half_triangles')
        expected = {'zero': 1, 'one': 1}
        assert converted == expected

        data = {'a': 0, 'b': 1}
        converted = rreg.convert(data, 'half_squares', 'half_triangles')
        expected = {'zero': 0.25, 'one': 0.75}
        assert converted == expected

    def test_convert_triangle_to_square(self, regions_half_squares, regions_half_triangles):
        rreg = RegionRegister()
        rreg.register(regions_half_squares)
        rreg.register(regions_half_triangles)

        data = {'zero': 1, 'one': 1}
        converted = rreg.convert(data, 'half_triangles', 'half_squares')
        expected = {'a': 1, 'b': 1}
        assert converted == expected

        data = {'zero': 0.25, 'one': 0.75}
        converted = rreg.convert(data, 'half_triangles', 'half_squares')
        expected = {'a': 0.375, 'b': 0.625}
        assert converted == expected



# -*- coding: utf-8 -*-
"""Read and parse the config for sector models
"""
import logging
import os

from .load import load
from .validate import (validate_initial_conditions, validate_input_spec,
                       validate_interventions, validate_output_spec)


class SectorModelReader(object):
    """Parses the configuration and input data for a sector model

    Arguments
    =========
    initial_config : dict
        Sector model details, sufficient to read the full config from a set of
        files. Must contain the following fields:

            "model_name"
                The name of the sector model, for reference within the
                system-of-systems model

            "model_path"
                The path to the python module file that contains an
                implementation of SectorModel

            "model_classname"
                The name of the class that implements SectorModel

            "model_config_dir"
                The root path of model config/data to use, which
                must contain inputs.yaml, outputs.yaml, time_intervals.yaml and
                regions.shp/regions.geojson

            "initial_conditions"
                List of files containing initial conditions

            "interventions"
                List of files containing interventions

    """
    def __init__(self, initial_config=None):
        self.logger = logging.getLogger(__name__)

        if initial_config is not None:
            self.model_name = initial_config["model_name"]
            self.model_path = initial_config["model_path"]
            self.model_classname = initial_config["model_classname"]
            self.model_config_dir = initial_config["model_config_dir"]
            self.initial_conditions_paths = initial_config["initial_conditions"]
            self.interventions_paths = initial_config["interventions"]
        else:
            self.model_name = None
            self.model_path = None
            self.model_classname = None
            self.model_config_dir = None
            self.initial_conditions_paths = None
            self.interventions_paths = None

        self.inputs = None
        self.outputs = None
        self.initial_conditions = None
        self.interventions = None

    def load(self):
        """Load and check all config
        """
        self.inputs = self.load_inputs()
        self.outputs = self.load_outputs()
        self.initial_conditions = self.load_initial_conditions()
        self.interventions = self.load_interventions()

    @property
    def data(self):
        """Expose all loaded config data

        Returns
        =======
        data : dict
            Model configuration data, with the following fields:
                "name": The name of the sector model, for reference within the
                system-of-systems model

                "path": The path to the python module file that contains an
                implementation of SectorModel

                "classname": The name of the class that implements SectorModel

                "inputs": A list of the inputs that this model requires

                "outputs": A list of the outputs that this model provides

                "time_intervals": A list of time intervals within a year that are
                represented by the model, each with reference to the model's
                internal identifier for timesteps

                "regions": A list of geographical regions used within the model, as
                objects with both geography and attributes

                "initial_conditions": A list of initial conditions required to set up
                the modelled system in the base year

                "interventions": A list of possible interventions that could be made
                in the modelled system

        """
        return {
            "name": self.model_name,
            "path": self.model_path,
            "classname": self.model_classname,
            "inputs": self.inputs,
            "outputs": self.outputs,
            "initial_conditions": self.initial_conditions,
            "interventions": self.interventions
        }

    def load_inputs(self):
        """Input spec is located in the ``data/<sectormodel>/inputs.yaml`` file

        """
        return self.load_io_metadata('inputs')

    def load_outputs(self):
        """Output spec is located in ``data/<sectormodel>/output.yaml`` file
        """
        return self.load_io_metadata('outputs')

    def load_io_metadata(self, inputs_or_outputs):
        """Load inputs or outputs, allowing missing or empty file
        """
        path = os.path.join(
            self.model_config_dir,
            '{}.yaml'.format(inputs_or_outputs))

        if not os.path.exists(path):
            msg = "No %s provided for '%s' model: %s not found"
            self.logger.warning(msg, inputs_or_outputs, self.model_name, path)
            data = []

        else:
            file_contents = load(path)

            if file_contents is None:
                self.logger.warning(
                    "No %s provided for '%s' model: %s was empty",
                    inputs_or_outputs,
                    self.model_name,
                    path)
                data = []
            else:
                data = file_contents

        if inputs_or_outputs == 'inputs':
            validate_input_spec(data, self.model_name)
        else:
            validate_output_spec(data, self.model_name)

        return data

    def load_initial_conditions(self):
        """Inital conditions are located in yaml files
        specified in sector model blocks in the sos model config
        """
        data = []
        for path in self.initial_conditions_paths:
            self.logger.debug("Loading initial conditions from %s", path)
            new_data = load(path)
            validate_initial_conditions(new_data, path)
            data.extend(new_data)
        return data

    def load_interventions(self):
        """Interventions are located in yaml files
        specified in sector model blocks in the sos model config
        """
        data = []
        paths = self.interventions_paths
        for path in paths:
            self.logger.debug("Loading interventions from %s", path)
            new_data = load(path)
            validate_interventions(new_data, path)
            data.extend(new_data)
        return data




from collections import OrderedDict
import numpy as np
from numpy.testing import assert_equal
from pytest import approx, fixture, raises
from smif.convert.interval import Interval, TimeIntervalRegister, TimeSeries


@fixture(scope='function')
def remap_months():
    """Remapping four representative months to months across the year

    In this case we have a model which represents the seasons through
    the year using one month for each season. We then map the four
    model seasons 1, 2, 3 & 4 onto the months throughout the year that
    they represent.

    The data will be presented to the model using the four time intervals,
    1, 2, 3 & 4. When converting to hours, the data will be replicated over
    the year.  When converting from hours to the model time intervals,
    data will be averaged and aggregated.

    """
    data = [{'id': '1', 'start': 'P0M', 'end': 'P1M'},
            {'id': '2', 'start': 'P1M', 'end': 'P2M'},
            {'id': '2', 'start': 'P2M', 'end': 'P3M'},
            {'id': '2', 'start': 'P3M', 'end': 'P4M'},
            {'id': '3', 'start': 'P4M', 'end': 'P5M'},
            {'id': '3', 'start': 'P5M', 'end': 'P6M'},
            {'id': '3', 'start': 'P6M', 'end': 'P7M'},
            {'id': '4', 'start': 'P7M', 'end': 'P8M'},
            {'id': '4', 'start': 'P8M', 'end': 'P9M'},
            {'id': '4', 'start': 'P9M', 'end': 'P10M'},
            {'id': '1', 'start': 'P10M', 'end': 'P11M'},
            {'id': '1', 'start': 'P11M', 'end': 'P12M'}]
    return data


@fixture(scope='function')
def data_remap():
    data = [{'id': '1', 'value': 30+31+31},
            {'id': '2', 'value': 28+31+30},
            {'id': '3', 'value': 31+31+30},
            {'id': '4', 'value': 30+31+31}]
    return data


@fixture(scope='function')
def expected_data_remap():
    data = [{'id': '1_0', 'value': 30.666666666},
            {'id': '1_1', 'value': 29.666666666},
            {'id': '1_2', 'value': 29.666666666},
            {'id': '1_3', 'value': 29.666666666},
            {'id': '1_4', 'value': 30.666666666},
            {'id': '1_5', 'value': 30.666666666},
            {'id': '1_6', 'value': 30.666666666},
            {'id': '1_7', 'value': 30.666666666},
            {'id': '1_8', 'value': 30.666666666},
            {'id': '1_9', 'value': 30.666666666},
            {'id': '1_10', 'value': 30.666666666},
            {'id': '1_11', 'value': 30.666666666}]
    return data


@fixture(scope='function')
def months():
    months = [{'id': '1_0', 'start': 'P0M', 'end': 'P1M'},
              {'id': '1_1', 'start': 'P1M', 'end': 'P2M'},
              {'id': '1_2', 'start': 'P2M', 'end': 'P3M'},
              {'id': '1_3', 'start': 'P3M', 'end': 'P4M'},
              {'id': '1_4', 'start': 'P4M', 'end': 'P5M'},
              {'id': '1_5', 'start': 'P5M', 'end': 'P6M'},
              {'id': '1_6', 'start': 'P6M', 'end': 'P7M'},
              {'id': '1_7', 'start': 'P7M', 'end': 'P8M'},
              {'id': '1_8', 'start': 'P8M', 'end': 'P9M'},
              {'id': '1_9', 'start': 'P9M', 'end': 'P10M'},
              {'id': '1_10', 'start': 'P10M', 'end': 'P11M'},
              {'id': '1_11', 'start': 'P11M', 'end': 'P12M'}]
    return months


@fixture(scope='function')
def monthly_data():
    """[31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    """
    data = [{'id': '1_0', 'value': 31},
            {'id': '1_1', 'value': 28},
            {'id': '1_2', 'value': 31},
            {'id': '1_3', 'value': 30},
            {'id': '1_4', 'value': 31},
            {'id': '1_5', 'value': 30},
            {'id': '1_6', 'value': 31},
            {'id': '1_7', 'value': 31},
            {'id': '1_8', 'value': 30},
            {'id': '1_9', 'value': 31},
            {'id': '1_10', 'value': 30},
            {'id': '1_11', 'value': 31}]
    return data


@fixture(scope='function')
def seasons():
    # NB "winter" is split into two pieces around the year end
    return [{'id': 'winter', 'start': 'P0M', 'end': 'P2M'},
            {'id': 'spring', 'start': 'P2M', 'end': 'P5M'},
            {'id': 'summer', 'start': 'P5M', 'end': 'P8M'},
            {'id': 'autumn', 'start': 'P8M', 'end': 'P11M'},
            {'id': 'winter', 'start': 'P11M', 'end': 'P12M'}]


@fixture(scope='function')
def twenty_four_hours():
    twenty_four_hours = \
        [{'id': '1_0', 'start': 'PT0H', 'end': 'PT1H'},
         {'id': '1_1', 'start': 'PT1H', 'end': 'PT2H'},
         {'id': '1_2', 'start': 'PT2H', 'end': 'PT3H'},
         {'id': '1_3', 'start': 'PT3H', 'end': 'PT4H'},
         {'id': '1_4', 'start': 'PT4H', 'end': 'PT5H'},
         {'id': '1_5', 'start': 'PT5H', 'end': 'PT6H'},
         {'id': '1_6', 'start': 'PT6H', 'end': 'PT7H'},
         {'id': '1_7', 'start': 'PT7H', 'end': 'PT8H'},
         {'id': '1_8', 'start': 'PT8H', 'end': 'PT9H'},
         {'id': '1_9', 'start': 'PT9H', 'end': 'PT10H'},
         {'id': '1_10', 'start': 'PT10H', 'end': 'PT11H'},
         {'id': '1_11', 'start': 'PT11H', 'end': 'PT12H'},
         {'id': '1_12', 'start': 'PT12H', 'end': 'PT13H'},
         {'id': '1_13', 'start': 'PT13H', 'end': 'PT14H'},
         {'id': '1_14', 'start': 'PT14H', 'end': 'PT15H'},
         {'id': '1_15', 'start': 'PT15H', 'end': 'PT16H'},
         {'id': '1_16', 'start': 'PT16H', 'end': 'PT17H'},
         {'id': '1_17', 'start': 'PT17H', 'end': 'PT18H'},
         {'id': '1_18', 'start': 'PT18H', 'end': 'PT19H'},
         {'id': '1_19', 'start': 'PT19H', 'end': 'PT20H'},
         {'id': '1_20', 'start': 'PT20H', 'end': 'PT21H'},
         {'id': '1_21', 'start': 'PT21H', 'end': 'PT22H'},
         {'id': '1_22', 'start': 'PT22H', 'end': 'PT23H'},
         {'id': '1_23', 'start': 'PT23H', 'end': 'PT24H'}]

    return twenty_four_hours


@fixture(scope='function')
def one_day():

    one_day = [{'id': 'one_day', 'start': 'P0D', 'end': 'P1D'}]

    return one_day


class TestInterval:

    def test_empty_interval_list(self):

        with raises(ValueError):
            Interval('test', [])

    def test_equality(self):
        a = Interval('test', ('PT0H', 'PT1H'))
        b = Interval('test', ('PT0H', 'PT1H'))
        c = Interval('another_test', ('PT0H', 'PT1H'))
        d = Interval('another_test', ('PT1H', 'PT2H'))

        assert a == b
        assert a != c
        assert b != c
        assert c != d

    def test_invalid_set_interval(self):
        interval = Interval('test', ('PT0H', 'PT1H'))
        with raises(ValueError) as excinfo:
            interval.interval = None
        msg = "A time interval must add either a single tuple or a list of tuples"
        assert msg in str(excinfo)

    def test_empty_interval_tuple(self):
        with raises(ValueError):
            Interval('test', ())

    def test_empty_interval_list_tuple(self):
        with raises(ValueError):
            Interval('test', [()])

    def test_load_interval(self):

        interval = Interval('test', ('PT0H', 'PT1H'))

        assert interval.name == 'test'
        assert interval.start == 'PT0H'
        assert interval.end == 'PT1H'

    def test_load_multiple_interval(self):

        interval = Interval('test', [('PT0H', 'PT1H'), ('PT1H', 'PT2H')])

        assert interval.name == 'test'
        assert interval.start == ['PT0H', 'PT1H']
        assert interval.end == ['PT1H', 'PT2H']

    def test_add_list_of_interval(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        interval.interval = [('PT1H', 'PT2H')]

        assert interval.name == 'test'
        assert interval.start == ['PT0H', 'PT1H']
        assert interval.end == ['PT1H', 'PT2H']

    def test_raise_error_load_illegal(self):
        with raises(ValueError):
            Interval('test', 'start', 'end')

    def test_convert_hour_to_hours(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        actual = interval._convert_to_hours(interval.start)
        expected = 0
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 1
        assert actual == expected

    def test_convert_month_to_hours(self):

        interval = Interval('test', ('P1M', 'P2M'))
        actual = interval._convert_to_hours(interval.start)
        expected = 744
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 1416
        assert actual == expected

    def test_convert_week_to_hours(self):

        interval = Interval('test', ('P2D', 'P3D'))
        actual = interval._convert_to_hours(interval.start)
        expected = 48
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 72
        assert actual == expected

    def test_to_hours_zero(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        actual = interval.to_hours()

        assert actual == [(0, 1)]

    def test_to_hours_month(self):

        interval = Interval('test', ('P2M', 'P3M'))
        actual = interval.to_hours()

        assert actual == [(1416, 2160)]

    def test_to_hours_list(self):

        interval = Interval('test', [('PT0H', 'PT1H'),
                                     ('PT2H', 'PT3H'),
                                     ('PT5H', 'PT7H')])
        actual = interval.to_hours()

        assert actual == [(0, 1), (2, 3), (5, 7)]

    def test_str_one_interval(self):
        interval = Interval('test', ('P2M', 'P3M'))
        actual = str(interval)
        assert actual == \
            "Interval 'test' maps to:\n  hour 1416 to hour 2160\n"

    def test_str_multiple_intervals(self):
        interval = Interval('test', [('PT0H', 'PT1H'),
                                     ('PT2H', 'PT3H'),
                                     ('PT5H', 'PT7H')])
        actual = str(interval)
        assert actual == \
            "Interval 'test' maps to:\n  hour 0 to hour 1\n  hour 2 to hour 3\n  "\
            "hour 5 to hour 7\n"

    def test_repr(self):
        interval = Interval('test', ('P2M', 'P3M'), 2011)
        actual = repr(interval)
        assert actual == "Interval('test', [('P2M', 'P3M')], base_year=2011)"

    def test_load_remap_timeslices(self):
        interval = Interval('1', [('P2M', 'P3M'),
                                  ('P3M', 'P4M'),
                                  ('P4M', 'P5M')])
        actual = interval.to_hours()
        assert actual == [(1416, 2160), (2160, 2880), (2880, 3624)]


class TestTimeSeries:

    def test_load_time_series(self, months):

        data = [{'id': '1_0', 'value': 1},
                {'id': '1_1', 'value': 1},
                {'id': '1_2', 'value': 1},
                {'id': '1_3', 'value': 1},
                {'id': '1_4', 'value': 1},
                {'id': '1_5', 'value': 1},
                {'id': '1_6', 'value': 1},
                {'id': '1_7', 'value': 1},
                {'id': '1_8', 'value': 1},
                {'id': '1_9', 'value': 1},
                {'id': '1_10', 'value': 1},
                {'id': '1_11', 'value': 1}]

        register = TimeIntervalRegister(2010)
        register.register(months, 'months')

        timeseries = TimeSeries(data)
        actual = timeseries.names
        expected = ['1_0', '1_1', '1_2', '1_3', '1_4', '1_5',
                    '1_6', '1_7', '1_8', '1_9', '1_10', '1_11']
        assert actual == expected

        actual = timeseries.values
        expected = [1] * 12
        assert actual == expected

        register._convert_to_hourly_buckets(timeseries, 'months')
        actual = timeseries.hourly_values

        month_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        month_hours = list(map(lambda x: x*24, month_days))
        expected_results = list(map(lambda x: 1/x, month_hours))

        start = 0
        for hours, expected in zip(month_hours, expected_results):
            expected_array = np.zeros(hours, dtype=np.float)
            expected_array.fill(expected)
            assert_equal(actual[start:start + hours], expected_array)
            start += hours


class TestTimeRegisterConversion:

    def test_raises_error_on_no_definition(self):

        register = TimeIntervalRegister()
        with raises(ValueError):
            register.get_intervals_in_set('blobby')

    def test_convert_from_month_to_seasons(self,
                                           months,
                                           seasons,
                                           monthly_data):

        data = monthly_data

        register = TimeIntervalRegister()
        register.register(months, 'months')
        register.register(seasons, 'seasons')

        timeseries = TimeSeries(data)

        actual = register.convert(timeseries, 'months', 'seasons')
        expected = [{'id': 'winter', 'value': 31. + 31 + 28},
                    {'id': 'spring', 'value': 31. + 30 + 31},
                    {'id': 'summer', 'value': 30. + 31 + 31},
                    {'id': 'autumn', 'value': 30. + 31 + 30}]

        for act, exp in zip(actual, expected):
            assert act['id'] == exp['id']
            assert act['value'] == approx(exp['value'])

    def test_convert_from_hour_to_day(self, twenty_four_hours, one_day):

        data = [{'id': '1_0', 'value': 1},
                {'id': '1_1', 'value': 1},
                {'id': '1_2', 'value': 1},
                {'id': '1_3', 'value': 1},
                {'id': '1_4', 'value': 1},
                {'id': '1_5', 'value': 1},
                {'id': '1_6', 'value': 1},
                {'id': '1_7', 'value': 1},
                {'id': '1_8', 'value': 1},
                {'id': '1_9', 'value': 1},
                {'id': '1_10', 'value': 1},
                {'id': '1_11', 'value': 1},
                {'id': '1_12', 'value': 1},
                {'id': '1_13', 'value': 1},
                {'id': '1_14', 'value': 1},
                {'id': '1_15', 'value': 1},
                {'id': '1_16', 'value': 1},
                {'id': '1_17', 'value': 1},
                {'id': '1_18', 'value': 1},
                {'id': '1_19', 'value': 1},
                {'id': '1_20', 'value': 1},
                {'id': '1_21', 'value': 1},
                {'id': '1_22', 'value': 1},
                {'id': '1_23', 'value': 1}]

        register = TimeIntervalRegister()
        register.register(twenty_four_hours, 'hourly_day')
        register.register(one_day, 'one_day')

        timeseries = TimeSeries(data)

        actual = register.convert(timeseries, 'hourly_day', 'one_day')
        expected = [{'id': 'one_day', 'value': 24}]

        assert actual == expected


class TestIntervalRegister:

    def test_interval_loads(self):
        """Pass a time-interval definition into the register

        """
        data = [{'id': '1_1',
                 'start': 'PT0H',
                 'end': 'PT1H'}]

        register = TimeIntervalRegister()
        register.register(data, 'energy_supply_hourly')
        assert register.interval_set_names == ['energy_supply_hourly']

        actual = register.get_intervals_in_set('energy_supply_hourly')

        element = Interval('1_1', ('PT0H', 'PT1H'), base_year=2010)
        expected = OrderedDict()
        expected['1_1'] = element

        assert actual == expected

    def test_interval_load_same_key(self, months):
        """Tests that interval sets with duplicate interval names can be
        loaded if they have different interval set names
        """
        register = TimeIntervalRegister()
        register.register(months, 'months_1')
        register.register(months, 'months_2')
        actual = register.get_intervals_in_set('months_1')
        expected = register.get_intervals_in_set('months_2')
        assert actual == expected

    def test_interval_load_duplicate_name_raises(self, months):
        """Tests that error is raised if a duplicate name is used
        for an interval set
        """
        register = TimeIntervalRegister()
        register.register(months, 'months_1')
        with raises(ValueError):
            register.register(months, 'months_1')

    def test_months_load(self, months):
        """Pass a monthly time-interval definition into the register

        """
        register = TimeIntervalRegister()
        register.register(months, 'months')

        actual = register.get_intervals_in_set('months')

        expected_names = \
            ['1_0', '1_1', '1_2', '1_3', '1_4', '1_5',
             '1_6', '1_7', '1_8', '1_9', '1_10', '1_11']

        expected = [Interval('1_0', ('P0M', 'P1M')),
                    Interval('1_1', ('P1M', 'P2M')),
                    Interval('1_2', ('P2M', 'P3M')),
                    Interval('1_3', ('P3M', 'P4M')),
                    Interval('1_4', ('P4M', 'P5M')),
                    Interval('1_5', ('P5M', 'P6M')),
                    Interval('1_6', ('P6M', 'P7M')),
                    Interval('1_7', ('P7M', 'P8M')),
                    Interval('1_8', ('P8M', 'P9M')),
                    Interval('1_9', ('P9M', 'P10M')),
                    Interval('1_10', ('P10M', 'P11M')),
                    Interval('1_11', ('P11M', 'P12M'))]

        for name, interval in zip(expected_names, expected):
            assert actual[name] == interval

    def test_remap_interval_load(self, remap_months):
        register = TimeIntervalRegister()
        register.register(remap_months, 'remap_months')

        actual = register.get_intervals_in_set('remap_months')

        expected_names = ['1', '2', '3', '4']

        expected = [Interval('1', [('P10M', 'P11M'),
                                   ('P11M', 'P12M'),
                                   ('P0M', 'P1M')]),
                    Interval('2', [('P1M', 'P2M'),
                                   ('P2M', 'P3M'),
                                   ('P3M', 'P4M')]),
                    Interval('3', [('P4M', 'P5M'),
                                   ('P5M', 'P6M'),
                                   ('P6M', 'P7M')]),
                    Interval('4', [('P7M', 'P8M'),
                                   ('P8M', 'P9M'),
                                   ('P9M', 'P10M')])]

        for name, interval in zip(expected_names, expected):
            assert actual[name] == interval


class TestRemapConvert:

    def test_remap_timeslices_to_months(self,
                                        months,
                                        expected_data_remap,
                                        remap_months,
                                        data_remap):
        """
        """
        timeslice_data = data_remap

        register = TimeIntervalRegister()
        register.register(months, 'months')
        register.register(remap_months, 'remap_months')

        timeseries = TimeSeries(timeslice_data)

        actual = register.convert(timeseries, 'remap_months', 'months')
        expected = expected_data_remap

        assert len(actual) == len(expected)

        for act, exp in zip(actual, expected):
            assert act['id'] == exp['id']
            assert act['value'] == approx(exp['value'])

    def test_remap_months_to_timeslices(self,
                                        months,
                                        monthly_data,
                                        remap_months,
                                        data_remap):
        """
        """
        timeslice_data = data_remap
        monthly_data = monthly_data

        register = TimeIntervalRegister()
        register.register(months, 'months')
        register.register(remap_months, 'remap_months')

        timeseries = TimeSeries(monthly_data)

        actual = register.convert(timeseries, 'months', 'remap_months')
        expected = timeslice_data

        assert len(actual) == len(expected)

        for act, exp in zip(actual, expected):
            assert act['id'] == exp['id']
            assert act['value'] == approx(exp['value'])


class TestValidation:

    def test_validate_get_hourly_array(self, remap_months):

        data = remap_months
        register = TimeIntervalRegister()
        register.register(data, 'remap_months')

        actual = register._get_hourly_array('remap_months')
        expected = np.ones(8760, dtype=np.int)
        assert_equal(actual, expected)

    def test_validate_intervals_passes(self, remap_months):

        data = remap_months
        register = TimeIntervalRegister()
        register.register(data, 'remap_months')

    def test_validate_intervals_fails(self, remap_months):

        data = remap_months
        data.append({'id': '5', 'start': 'PT0H', 'end': 'PT1H'})
        register = TimeIntervalRegister()
        with raises(ValueError) as excinfo:
            register.register(data, 'remap_months')
        assert "Duplicate entry for hour 0 in interval set remap_months." in str(excinfo.value)

    def test_time_interval_start_before_end(get_time_intervals):
        with raises(ValueError) as excinfo:
            Interval('backwards', ('P1Y', 'P3M'))
        assert "A time interval must not end before it starts" in str(excinfo)

        interval = Interval('starts_ok', ('P0Y', 'P1M'))
        with raises(ValueError) as excinfo:
            interval.interval = ('P2M', 'P1M')
        assert "A time interval must not end before it starts" in str(excinfo)



"""Tests the ModelInputs class

"""
from pytest import fixture, raises
from smif.parameters import ModelParameters


class TestModelInputs:
    """Given a list of parameter dicts of the format::

    [
            {
                'name': 'macguffins produced',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

    Return a name tuple of dependencies

    """
    def test_one_dependency(self, one_dependency):
        inputs = ModelParameters(one_dependency)

        actual = inputs.names
        expected = ['macguffins produced']
        assert actual == expected

        actual = inputs.spatial_resolutions
        expected = ['LSOA']
        assert actual == expected
        actual = inputs.temporal_resolutions
        expected = ['annual']
        assert actual == expected

        actual = len(inputs)
        expected = 1
        assert actual == expected


@fixture(scope='function')
def two_output_metrics():
    """Returns a model output dictionary with two metrics
    """
    outputs = [
        {'name': 'total_cost',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'},
        {'name': 'water_demand',
            'spatial_resolution': 'watershed',
            'temporal_resolution': 'daily'}
        ]
    return outputs


class TestModelOutputs:
    """Given a list of dicts from ``outputs.yaml``, return a list of names::

        metrics:
        - name: total_cost
        - name: water_demand

    """

    def test_model_outputs(self, two_output_metrics):

        outputs = ModelParameters(two_output_metrics)
        actual = [x for x in outputs.parameters]
        names = ['total_cost', 'water_demand']
        areas = ['LSOA', 'watershed']
        intervals = ['annual', 'daily']

        for actual, name, area, inter in zip(outputs.parameters,
                                             names,
                                             areas,
                                             intervals):
            assert actual.name == name
            assert actual.spatial_resolution == area
            assert actual.temporal_resolution == inter

    def test_get_spatial_property(self, two_output_metrics):

        outputs = ModelParameters(two_output_metrics)

        actual = outputs.get_spatial_res('total_cost')
        expected = 'LSOA'
        assert actual == expected

        actual = outputs.get_spatial_res('water_demand')
        expected = 'watershed'
        assert actual == expected

        with raises(ValueError) as ex:
            outputs.get_spatial_res('missing')
        assert "No output found for name 'missing'" in str(ex.value)

    def test_get_temporal_property(self, two_output_metrics):

        outputs = ModelParameters(two_output_metrics)

        actual = outputs.get_temporal_res('total_cost')
        expected = 'annual'
        assert actual == expected

        actual = outputs.get_temporal_res('water_demand')
        expected = 'daily'
        assert actual == expected

        with raises(ValueError) as ex:
            outputs.get_temporal_res('missing')
        assert "No output found for name 'missing'" in str(ex.value)
